import os
import re
from copy import copy
import traceback
from decorator import decorator
from functools import wraps
from itertools import product
import logging
from collections import OrderedDict
import numpy as np

from pymake import ExpSpace, get_pymake_settings
from pymake.util.utils import colored, basestring, make_path, hash_objects

from tabulate import tabulate



''' Sanbox Base class for expe execution.
'''



### Todo:
#  try to read the decorator that were called for _[post|pre]process
# * if @plot then display
# * if @tabulate then ...
class ExpeFormat(object):
    ''' A Base class for processing individuals experiments (**expe**).

        Notes
        -----
        The following attribute have a special meaning when subclassing:
            * _default_expe : is updated by each single expe.

    '''

    log = logging.getLogger('root')
    _logfile = False # for external integration @deprcated ?

    def __init__(self, pt, expe, expdesign, gramexp):
        ''' Sandbox class for scripts/actions.

            Parameters
            ----------
            pt: int
                Positional indicator for current run
            expe: ExpSpace
                Current spec
            expdesign: ExpDesign
                Current design class
            gramexp: Global pmk object

            Methods
            -------
            log_expe:
                overwrite the header message before each run.
            is_first_expe:
                return true if the current run is the first.
            is_last_expe:
                return true if the current run is the last.
            load_frontend:
                load the data frontend for the current expe.
            load_model:
                the model for the current expe.

            Decorator
            --------
            plot:
                plot control on expe
            table:
                table control on expe

        '''

        self._expdesign = expdesign

        # Global
        self.expe_size = len(gramexp)
        self.gramexp = gramexp
        self.D = self.gramexp.D # Global conteneur
        # Local
        self.pt = pt
        self.expe = expe

        # Plot utils
        from pymake.plot import _linestyle, _markers, _colors
        self.linestyles = _linestyle.copy()
        self.markers = _markers.copy()
        self.colors = _colors.copy()

        # to exploit / Vizu
        self._it = pt['expe']
        self.corpus_pos = pt.get('corpus')
        self.model_pos = pt.get('model')
        self.output_path = self.expe['_output_path']
        self.input_path = self.expe['_input_path']

        if expe.get('_expe_silent'):
            self.log_silent()
        else:
            self.log.info('-'*10)
            self.log.info(''.join([colored('Expe %d/%d', 'red'), ' : ', self.log_expe()]) % (
                self._it+1, self.expe_size,
            ))
            self.log.info('-'*10)

        self._set_csv_typo()

    def _set_csv_typo(self):
        csv_typo = self.expe.get('_csv_typo')
        if csv_typo is None:
            if 'model' in self.expe:
                from pymake.frontend.manager import ModelManager
                model = ModelManager.from_expe(self.expe)
                csv_typo = getattr(model, '_csv_typo', None)
                try:
                    from pymake.frontend.manager import ModelManager
                    model = ModelManager.from_expe(self.expe)
                    csv_typo = getattr(model, '_csv_typo', None)
                except Exception as e:
                    self.log.debug('No csv_typo spec found...: %s' % e)
                    csv_typo = None

        self._csv_typo = csv_typo

    def log_expe(self):
        expe = self.expe
        msg = '%s -- %s -- N=%s -- K=%s' % (self.specname(expe.get('corpus')),
                                            self.specname(expe.get('model')),
                                            expe.get('N'), expe.get('K'))
        return msg

    def log_silent(self):
        if self.is_first_expe():
            print()

        prefix = 'Computing'
        n_it = self._it +1
        n_total = self.expe_size
        # Normalize
        n_it_norm = 2*42 * n_it // n_total

        progress= n_it_norm * '='  + (2*42-n_it_norm) * ' '
        print('\r%s: [%s>] %s/%s' % (prefix, progress, n_it, n_total), end = '\r')

        if self.is_last_expe():
            print()
            print()

    def is_first_expe(self):
        if 0 == self.pt['expe']:
            return True
        else:
            return False

    def is_last_expe(self):
        if self.expe_size - 1 == self.pt['expe']:
            return True
        else:
            return False

    def spec_from_expe(self, spec_map=None):
        ''' Return a sub dict from expe, from spec_map. '''
        if spec_map is None:
            spec_map = {}

        spec = {}
        for k, v in spec_map.items():
            if v in self.expe:
                spec[k] = self.expe[v]

        return spec

    def get_data_path(self):
        path = get_pymake_settings('project_data')
        path = os.path.join(path, '')
        return path

    def getenv(self, a=None):
        return self.gramexp.getenv(a)

    def get_expset(self, param):
        return self.gramexp.get_set(param)

    def get_figs(self):
        return list(self.gramexp._figs.values())

    def get_tables(self):
        return list(self.gramexp._tables.values())

    def get_current_group(self):
        return self._file_part(sorted([str(self.expe.get(g)) for g in self._groups]), sep='-')

    def get_current_frame(self):
        group = self.get_current_group()
        frame = self.gramexp._figs[group]
        return frame


    def get_description(self, full=False):
        if full:
            return '/'.join((self.expe._refdir, os.path.basename(self.output_path)))
        else:
            return os.path.basename(self.output_path)

    @classmethod
    def tabulate(cls, *args, **kwargs):
        return tabulate(*args, **kwargs)

    @classmethod
    def display(cls, conf):
        import matplotlib.pyplot as plt
        block = not conf.get('_no_block_plot', False)
        plt.show(block=block)

    @staticmethod
    @decorator
    def plot_simple(fun, *args, **kwargs):
        # @obsolete ? (no block in save, this is it?
        import matplotlib.pyplot as plt
        self = args[0]
        expe = self.expe
        kernel = fun(*args, **kwargs)
        if '_no_block_plot' in expe and getattr(self, 'noplot', False) is not True:
            plt.show(block=expe._no_block_plot)
        return kernel

    @staticmethod
    def expe_repeat(fun):

        def wrapper(self, *args, **kwargs):

            repeats = self.get_expset('_repeat')
            if len(repeats)>1:
                kwargs['repeat'] = True

            res = fun(self, *args, **kwargs)

            return res
        return wrapper


    @staticmethod
    def raw_plot(*groups, **_kwargs):
        ''' If no argument, simple plot => @obsolete?
            If arguments (in decorator) :
                * [0] : group figure by this
                * [1] : key for id (title and filename)
        '''
        if groups and len(groups[1:]) == 0 and callable(groups[0]):
            # decorator whithout arguments
            return ExpeFormat.plot_simple(groups[0])

        def decorator(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                import matplotlib.pyplot as plt
                from pymake.plot import _linestyle, _markers
                if len(groups) == 0:
                    group = '_id_expe'
                else:
                    group = groups[0]
                self = args[0]
                expe = self.expe
                discr_args = []
                if len(args) > 1:
                    discr_args = args[1].split('/')

                # Init Figs Sink
                if not hasattr(self.gramexp, '_figs'):
                    figs = dict()
                    for c in self.gramexp.get_set(group):
                        figs[c] = ExpSpace()
                        figs[c].group = c
                        figs[c].fig = plt.figure()
                        figs[c].linestyle = _linestyle.copy()
                        figs[c].markers = _markers.copy()

                    self.gramexp._figs = figs

                frame = self.gramexp._figs[expe[group]]
                frame.ax = lambda: frame.fig.gca()

                kernel = fun(self, frame, *args[1:])

                # Set title and filename
                title = ' '.join('{{{0}}}'.format(w) for w in groups).format(**self.specname(expe))
                frame.base = '%s_%s' % (fun.__name__, title.replace(' ', '_'))
                frame.args = discr_args
                if 'title' in frame:
                    # if multipolot
                    #plt.suptitle(frame.title)
                    frame.ax().set_title(frame.title)
                else:
                    frame.ax().set_title(title)

                # Save on last call
                if self._it == self.expe_size -1:
                    if expe._write:
                        self.write_frames(self.gramexp._figs)

                return kernel
            return wrapper

        return decorator

    @staticmethod
    def plot(*a, **b):
        ''' Doc todo (plot alignement...) '''

        def decorator(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                import matplotlib.pyplot as plt
                from pymake.plot import _linestyle, _markers

                self = args[0]
                discr_args = [] # discriminant keys (to distinguish filename)
                if len(args) == 1:
                    groups = ['corpus']
                    attribute = '_entropy'
                else:
                    groups = args[1].split(':')
                    if len(groups) == 1:
                        attribute = groups[0]
                        groups = None
                    else:
                        attribute = groups.pop(-1)
                        groups = groups[0]
                    if len(args) > 2:
                        discr_args = args[2].split('/')

                if groups:
                    groups = groups.split('/')
                    self._groups = groups
                    ggroup = self.get_current_group()
                else:
                    ggroup = None

                # Init Figs Sink
                if not hasattr(self.gramexp, '_figs'):
                    figs = dict()
                    if groups:
                        gset = product(*filter(None, [self.gramexp.get_set(g) for g in groups]))
                    else:
                        gset = [None]

                    for g in gset:
                        gg = '-'.join(sorted(map(str,g))) if g else None
                        figs[gg] = ExpSpace()
                        figs[gg].group = g
                        figs[gg].fig = plt.figure()
                        figs[gg].linestyle = _linestyle.copy()
                        figs[gg].markers = _markers.copy()

                    self.gramexp._figs = figs

                frame = self.gramexp._figs[ggroup]
                frame.ax = lambda: frame.fig.gca()

                kernel = fun(self, frame, attribute)

                # Set title and filename
                if groups and self.expe.get(groups[0]):
                    #title = ' '.join('{{{0}}}'.format(w) for w in groups).format(**self.specname(self.expe))
                    ctitle = tuple(filter(None,map(lambda x:self.specname(self.expe.get(x, x)), groups)))
                    s = '_'.join(['%s'] * len(ctitle))
                    title = s % ctitle
                else:
                    title = ' '.join(self.gramexp.get_nounique_keys())
                    if not title:
                        title = '%s %s' % tuple(map(lambda x:self.expe.get(x, x), ['corpus', 'model']))

                frame.base = '%s_%s' % (fun.__name__, attribute)
                frame.args = discr_args

                if 'fig_xaxis' in frame or self.expe.get('fig_xaxis'):
                    xaxis_name = frame.get('xaxis', self.expe.get('fig_xaxis'))
                    frame.ax().set_xlabel(xaxis_name)
                else:
                    frame.ax().set_xlabel('iterations')

                if 'fig_yaxis' in frame or self.expe.get('fig_yaxis'):
                    yaxis_name = frame.get('yaxis', expe.get('fig_yaxis'))
                    frame.ax().set_ylabel(yaxis_name)
                else:
                    frame.ax().set_ylabel(attribute)


                if 'title_size' in self.expe:
                    ts =  float(self.expe['title_size'])
                else:
                    ts = 15

                if 'title' in frame:
                    plt.suptitle(frame.title, fontsize=ts)
                else:
                    frame.ax().set_title(title, fontsize=ts)

                if 'ticks_size' in self.expe:
                     plt.xticks(fontsize=float((self.expe['ticks_size'])))

                # Save on last call
                if self._it == self.expe_size -1:
                    if self.expe._write:
                        self.write_frames(self.gramexp._figs)

                return kernel
            return wrapper

        return decorator

    @staticmethod
    def table(*a, **b):
        ''' Doc todo (plot alignement...) '''

        def decorator(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                self = args[0]
                discr_args = [] # discriminant keys (to distinguish filename)
                if len(args) == 1:
                    x, y, z = 'corpus', 'model', '_entropy'
                else:
                    x, y, z = args[1].split(':')
                    if len(args) > 2:
                        if ',' in args[2]:
                            split_sep = '|'
                            discr_args = args[2].split(',')
                        else:
                            split_sep = '/'
                            discr_args = args[2].split('/')

                        kwargs['split_sep'] = split_sep

                if discr_args:
                    groups = discr_args
                    self._groups = groups
                    # or None if args not in expe (tex option...)
                    ggroup = self.get_current_group() or None
                else:
                    groups = None
                    ggroup = None

                if not hasattr(self.gramexp, '_tables'):
                # self.is_first_expe

                    tablefmt_ext = dict(simple='md', latex='tex')
                    tablefmt = 'latex' if 'tex' in discr_args else 'simple'

                    tables = OrderedDict()
                    if groups:
                        gset = product(*filter(None, [self.gramexp.get_set(g) for g in groups]))
                    else:
                        gset = [None]

                    for g in sorted(gset):
                        gg = '-'.join(sorted(map(str,g))) if g else None
                        tables[gg] = ExpSpace()
                        _z = z.split('-')
                        array, floc = self.gramexp.get_array_loc(x, y, _z, repeat=kwargs.get('repeat'))
                        tables[gg].name = gg
                        tables[gg].array = array
                        tables[gg].floc = floc
                        tables[gg].x = x
                        tables[gg].y = y
                        tables[gg].z = _z
                        tables[gg].headers = self.specname(self.gramexp.get_set(y))
                        tables[gg].column = self.specname(self.gramexp.get_set(x))
                        tables[gg].floatfmt = '.3f'
                        tables[gg].tablefmt = tablefmt
                        tables[gg].ext = tablefmt_ext[tablefmt]
                        tables[gg].args = discr_args
                        tables[gg].kwargs = kwargs
                        tables[gg].base = '_'.join((fun.__name__,
                                                str(self.expe[x]),
                                                str(self.expe[y]))),

                    self.gramexp._tables = tables

                frame = self.gramexp._tables[ggroup]
                floc = frame.floc
                if kwargs.get('repeat'):
                    repeat_pos = self.pt['_repeat']
                    array = frame.array[repeat_pos]
                else:
                    array = frame.array

                for z in frame.z:
                    kernel = fun(self, array, floc, x, y, z, *discr_args)

                if self._it == self.expe_size -1:
                    tables = []

                    self.aggregate_tables(**kwargs)
                    self.decompose_tables(**kwargs)

                    self.log_tables(**kwargs)

                    if self.expe._write:
                        self.write_frames(self.gramexp._tables)

                return kernel
            return wrapper

        return decorator


    def aggregate_tables(self, **kwargs):
        ''' Group table if separator is "|" '''

        split_sep = kwargs.get('split_sep')

        if split_sep == '|':
            titles = list(self.gramexp._tables.keys())
            tables = list(self.gramexp._tables.values())
            title = '|'.join(titles)

            # Assume all columns are the same.
            t0 = tables[0]
            headers = len(titles) * t0.headers
            rp = 1 if 'repeat' in kwargs else 0
            new_array = np.concatenate([tt.array for tt in tables], 1+rp)
            t0.array = new_array
            t0.headers = headers

            self.gramexp._tables.clear()
            self.gramexp._tables[title] = t0


    def decompose_tables(self, **kwargs):

        if kwargs.get('repeat'):
            for ggroup in list(self.gramexp._tables):
                _table = self.gramexp._tables.pop(ggroup)
                array = _table.array
                for zpos, z in enumerate(_table.z):

                    fop = np.max if 'rmax' in _table.args else np.mean
                    fop = np.min if 'rmin' in _table.args else fop

                    # Mean and standard deviation
                    table = array[:,:,:,zpos]
                    mtable = self.highlight_table(np.around(fop(table, 0), decimals=3))
                    vtable = np.around(table.std(0), decimals=3)
                    table_mean = np.char.array(mtable).astype("|S42")
                    table_std = np.char.array(vtable).astype("|S42")
                    arr = table_mean + b' pm ' + table_std

                    new_table = _table.copy()
                    new_table.array = arr
                    ngg = z +'-'+ ggroup if ggroup else z
                    self.gramexp._tables[ngg] = new_table

        else:
            for ggroup in list(self.gramexp._tables):
                _table = self.gramexp._tables.pop(ggroup)
                array = _table.array
                for zpos, z in enumerate(_table.z):
                    arr = self.highlight_table(array[:,:,zpos])

                    new_table = _table.copy()
                    new_table.array = arr
                    ngg = z +'-'+ ggroup if ggroup else z
                    self.gramexp._tables[ngg] = new_table

    def log_tables(self, **kwargs):

        for _title, _table in self.gramexp._tables.items():

            arr = _table.array.astype(str)
            table = np.column_stack((_table.column, arr))
            Table = tabulate(table, headers=_table.headers, tablefmt=_table.tablefmt, floatfmt=_table.floatfmt)

            print(colored('\n%s Table:'%(_title), 'green'))
            print(Table)

    @staticmethod
    def _file_part(group, sep='_'):
        part = [str(e) for e in group if (e != None and e != 'None')]
        part = sep.join(part)
        return part

    def highlight_table(self, array, highlight_dim=1):
        hack_float = np.vectorize(lambda x : '{:.3f}'.format(float(x)))
        table = np.char.array(hack_float(array), itemsize=42)
        # vectorize
        for i, col in enumerate(array.argmax(1)):
            table[i, col] = colored(table[i, col], 'bold')
        for i, col in enumerate(array.argmin(1)):
            table[i, col] = colored(table[i, col], 'magenta')

        return table


    def write_frames(self, frames, base='', suffix='', args=''):
        expe = self.formatName(self.expe)

        if isinstance(frames, str):
            frames = [frames]

        if type(frames) is list:
            if base:
                base = self.specname(base)
            if args:
                s = '_'.join(['%s'] * len(args))
                args = s % tuple(map(lambda x:expe.get(x, x), args))
            for i, f in enumerate(frames):
                idi = str(i) if len(frames) > 1 else None
                fn = self._file_part([base, args, suffix, idi])
                fn = self.full_fig_path(fn)
                self._kernel_write(f, fn)
        elif issubclass(type(frames), dict):
            for c, f in frames.items():
                base = f.get('base') or base
                args = f.get('args') or args
                if base:
                    base = self.specname(base)
                if args:
                    s = '_'.join(['%s'] * len(args))
                    args = s % tuple(map(lambda x:expe.get(x, x), args))
                fn = self._file_part([self.specname(c), base, args, suffix])
                fn = self.full_fig_path(fn)
                self._kernel_write(f, fn, title=c)
        else:
            print('Error : type of Frame unknow, passing: %s' % type(frame))


    def _kernel_write(self, frame, fn, title=None):

        if hasattr(frame, 'get'):
            ext = frame.get('ext')
        else:
            ext = None

        if isinstance(frame, dict):
            if 'fig' in frame:
                ext = ext or 'pdf'
                fn = fn +'.'+ ext
                print('Writing frame: %s' % fn)
                #frame.fig.tight_layout() # works better in parameter
                frame['fig'].savefig(fn, bbox_inches='tight') # pad_inches=-1
            elif '_table' in frame:
                ext = ext or 'md'
                fn = fn +'.'+ ext
                print('Writing frame: %s' % fn)
                caption = '\caption{{{title}}}\n'

                arr = frame['array'].astype(str)
                table = np.dstack((frame['column'], arr))
                Table = tabulate(table, headers=frame['headers'], tablefmt=frame['tablefmt'], floatfmt=frame['floatfmt'])

                with open(fn, 'w') as _f:
                    if  title:
                        _f.write(caption.format(title=title))
                    _f.write(Table+'\n')

        elif isinstance(frame, str):
            ext = ext or 'md'
            fn = fn +'.'+ ext
            print('Writing frame: %s' % fn)
            with open(fn, 'w') as _f:
                _f.write(frame)
        else:
            # assume figure
            ext = ext or 'pdf'
            fn = fn +'.'+ ext
            print('Writing frame: %s' % fn)
            frame.savefig(fn, bbox_inches='tight')

    def specname(self, n):
        #return self._expdesign._name(n).lower().replace(' ', '')
        return self._expdesign._name(n)

    def full_fig_path(self, fn):
        figs_path = get_pymake_settings('project_figs')
        path = os.path.join(figs_path, self.expe.get('_refdir',''),  self.specname(fn))
        make_path(path)
        return path

    def formatName(self, expe):
        expe = copy(expe)
        for k, v in expe.items():
            if isinstance(v, basestring):
                nn = self.specname(v)
            else:
                nn = v
            setattr(expe, k, nn)
        return expe




    @classmethod
    def _preprocess_(cls, gramexp):
        ''' This method has **write** access to Gramexp

            Notes
            -----
            Called once before running expe.
        '''

        # Update defautl settings of  gramexp
        # given _default_expe
        gramexp.update_default_expe(cls)

        # Put a valid expe a the end.
        gramexp.reorder_firstnonvalid()

        if not gramexp._conf.get('simulate'):
            cls.log.info(gramexp.exptable())

        # Global container shared by all expe
        # running in the sandbox.
        gramexp.D = ExpSpace()

        return

    @classmethod
    def _postprocess_(cls, gramexp):
        '''
            Notes
            -----
            Called once after all expe are finished.
        '''
        cls.display(gramexp._conf)
        return

    def _expe_preprocess(self):
        ''' system preprocess '''
        # setup seed ?
        # setup branch ?
        # setup description ?

        if hasattr(self, '_preprocess'):
            self._preprocess()

    def _expe_postprocess(self):
        ''' system postprocess '''

        if self.expe.get('_write'):

            # @improve: In ModelManager ?
            self.clear_fitfile()
            if hasattr(self, 'model') and hasattr(self.model, 'save'):
                if hasattr(self.model, 'write_current_state'):
                    self.model.save()

        if hasattr(self, '_postprocess'):
            self._postprocess()

    def _preprocess(self):
        ''' user defined preprocess '''
        # heere, do a decorator ?
        pass

    def _postprocess(self):
        ''' user defined postprocess '''
        # heere, do a decorator ?
        pass

    def __call__(self):
        raise NotImplementedError


    def _extract_csv_sample(self, model):
        ''' extract data in model from variable name declared in {self.scv_typo}.
            The iterable build constitute the csv-like line, for **one iteration**, to be written in the outputfile.

            Spec : (Make a grammar dude !)
                * Each name in the **typo** should be accesible (gettable) in the model/module class, at fit time,
                * a '{}' symbol means that this is a list.
                * a 'x[y]' symbole that the dict value is requested,
                * if there is a list '{}', the size of the list should be **put just before it**,
                * if there is several list next each other, they should have the same size.

                @debug: raise a sprcialerror from this class,
                    instead of duplicate the exception print in _fotmat_line_out.

                @todo use _fmt if given.
        '''
        line = []
        for o in self._csv_typo.split():
            if o.startswith('{'): # is a list
                obj = o[1:-1]
                brak_pt = obj.find('[')
                if brak_pt != -1: # assume  nested dictionary
                    obj, key = obj[:brak_pt], obj[brak_pt+1:-1]
                    try:
                        values = [str(elt[key]) for elt in getattr(model, obj).values()]
                    except (KeyError, AttributeError) as e:
                        values = self.format_error(model, o)

                else : # assume list
                    try:
                        values = [str(elt) for elt in getattr(model, obj)]
                    except (KeyError, AttributeError) as e:
                        values = self.format_error(model, o)
            else: # is atomic ie can be converted to string.
                try: values = str(getattr(model, o))
                except (KeyError, AttributeError) as e: values = self.format_error(model, o)


            if isinstance(values, list):
                line.extend(values)
            else:
                line.append(values)

        return line

    def format_error(self, model, o):
        traceback.print_exc()
        print('\n')
        self.log.critical("expe setting ${_format} is probably wrong !")
        self.log.error("model `%s' do not contains one of object: %s" % (str(model), o))
        print('Continue...')
        #os._exit(2)
        return 'None'

    def init_fitfile(self):
        ''' Create the file to save the iterations state of a model. '''
        self._samples = []
        os.makedirs(os.path.dirname(self.output_path), exist_ok=True)
        self.fname_i = self.output_path + '.inf'
        if not '_csv_typo' in self.expe:
            self.log.debug('No _csv_typo, for this model %s, no inference file...'%(self.expe.get('model')))
        else:
            self._fitit_f = open(self.fname_i, 'wb')
            self._fitit_f.write(('#' + self._csv_typo + '\n').encode('utf8'))


    def clear_fitfile(self):
        ''' Write remaining data and close the file. '''

        if not hasattr(self, '_fitit_f'):
            return

        if hasattr(self, '_samples') and self._samples:
            self._write_some(self._fitit_f, None)

        end_line = '# terminated\n'
        self._fitit_f.write(end_line.encode('utf8'))

        self._fitit_f.close()


    def load_some(self, filename=None, iter_max=None, comments=('#','%')):
        ''' Load data from file according that each line
            respect the format in {self._csv_typo}.
        '''
        if filename is None:
            filename = self.output_path + '.inf'

        if not os.path.exists(filename):
            return None

        with open(filename) as f:
            data = f.read()

        data = filter(None, data.split('\n'))
        if iter_max:
            data = data[:iter_max]
        # Ignore Comments
        data = [re.sub("\s\s+" , " ", x.strip()).split() for l,x in enumerate(data) if not x.startswith(comments)]

        # Grammar dude ?
        col_typo = self._csv_typo.split()
        array = []
        last_elt_size = None
        _pos_failed = set()
        # Format the data in a list of dict by entry
        for data_line in data:
            line = {}
            offset = 0
            for true_pos in range(len(data_line)):
                pos = true_pos + offset
                if pos >= len(data_line):
                    break

                try:
                    o = col_typo[true_pos]
                except:
                    #print(col_typo)
                    #print(data_line)
                    #print(true_pos)
                    if true_pos not in _pos_failed:
                        self.log.warning('No value found for pos: %d' % true_pos)
                        _pos_failed.add(true_pos)
                    continue

                if data_line[0] in comments:
                    break
                elif o.startswith('{'): # is a list
                    obj = o[1:-1]
                    brak_pt = obj.find('[')
                    if brak_pt != -1: # assume  nested dictionary
                        obj, key = obj[:brak_pt], obj[brak_pt+1:-1]
                        newkey = '.'.join((obj, key))
                        values = data_line[pos:pos+last_elt_size]
                        line[newkey] = values
                    else : # assume list
                        values = data_line[pos:pos+last_elt_size]
                        line[obj] = values
                    offset += last_elt_size-1
                else:
                    line[o] = data_line[pos]
                    if str.isdigit(line[o]):
                        # size of the list behind.
                        last_elt_size = int(line[o])

            array.append(line)


        # Format the array of dict to a dict by entry function of iterations
        data = {}
        for line in array:
            for k, v in line.items():
                l = data.get(k, [])
                l.append(v)
                data[k] = l

        return data


    def _write_some(self, _f, samples, buff=20):
        ''' Write data with buffer manager
            * lines are formatted as {self.csv_typo}
            * output file is {self._f}
        '''
        #fmt = self.fmt

        if samples is None:
            buff=1
        else:
            self._samples.append(samples)

        if len(self._samples) >= buff:
            #samples = np.array(self._samples)
            samples = self._samples
            #np.savetxt(f, samples, fmt=str(fmt))
            for line in samples:
                # @debug manage float .4f !
                line = ' '.join(line)+'\n'
                _f.write(line.encode('utf8'))

            _f.flush()
            self._samples = []


    def write_current_state(self, model):
        ''' push the current state of a model in the output file. '''
        samples = self._extract_csv_sample(model)
        self._write_some(self._fitit_f, samples)


    def configure_model(self, model):
        ''' Configure Model:
            * [warning] it removes existing [expe_path].inf file
        '''

        self.model = model

        # Inject the inline-writing method
        setattr(model, 'write_current_state', self.write_current_state)

        if self.expe.get('_write'):
            self.init_fitfile()

        # Could configure frontend/data path or more also here ?
        return

    # frontend params is deprecated and will be removed soon...
    def load_model(self, frontend=None, load=False):
        ''' :load: boolean. Load from **preprocess** file is true else
                            it is a raw loading.
        '''
        from pymake.frontend.manager import ModelManager

        self.model = ModelManager.from_expe(self.expe, frontend=frontend, load=load)
        if load is False:
            self.configure_model(self.model)

        return self.model

    def load_frontend(self):
        ''' See -nld and -sld option for control over load/save status
            of frontend data.
        '''
        from pymake.frontend.manager import FrontendManager

        frontend = FrontendManager.load(self.expe)
        return frontend







import abc
import numpy as np
import scipy as sp
import scipy.stats as stats
import math
import logging
lgg = logging.getLogger('root')

""" This code was modified from the code
originally written by Zhai Ke (kzhai@umd.edu)."""

class IBP(object):
    __metaclass__ = abc.ABCMeta

    """
    @param alpha_hyper_parameter: hyper-parameter for alpha sampling, a tuple defining the parameter for an inverse gamma distribution
    @param metropolis_hastings_k_new: a boolean variable, set to true if we use metropolis hasting to estimate K_new, otherwise use truncated gibbs sampling """
    def __init__(self, #real_valued_latent_feature=True,
                 alpha_hyper_parameter='',
                 metropolis_hastings_k_new=True):
        # initialize the hyper-parameter for sampling _alpha
        # a value of None is a gentle way to say "do not sampling _alpha"
        if alpha_hyper_parameter.startswith('auto'):
            self._alpha_hyper_parameter = (1., 1.)
        else:
            self._alpha_hyper_parameter = False

        self._metropolis_hastings_k_new = metropolis_hastings_k_new

        self._x_title = "X-matrix-"
        self._z_title = "Z-matrix-"
        self._hyper_parameter_vector_title = "Hyper-parameter-vector-"
        # to remove
        #super(IBP, self).__init__()

    """
    @param data: a NxD np data matrix
    @param alpha: IBP hyper parameter
    @param initializ_Z: seeded Z matrix """
    #@abc.abstractmethod
    def _initialize(self, frontend, alpha=1.0, initial_Z=None, KK=None):
        self._alpha = alpha

        # Data matrix
        #self._Y = self.center_data(data)
        self._Y = frontend.data_ma

        # Binary case
        Yd = self._Y.data
        Yd[Yd <= 0 ] = -1
        Yd[Yd > 0 ] = 1
        (self._N, self._D) = self._Y.shape

        if(initial_Z == None):
            # initialize Z from IBP(alpha)
            if KK is None:
                self._Z = self.initialize_Z()
            else:
                self._Z = self.initialize_Z_debug(KK=KK)
        else:
            self._Z = initial_Z

        assert(self._Z.shape[0] == self._N)

        # make sure Z matrix is a binary matrix
        assert(self._Z.dtype == np.int)
        assert(self._Z.max() == 1 and self._Z.min() == 0)

        # record down the number of features
        self._K = self._Z.shape[1]

        return


    """
    initialize latent feature appearance matrix Z according to IBP(alpha) """
    def initialize_Z(self, N=None, alpha=None, left_ordered=True):
        if N is None:
            N = self._N
        if alpha is None:
            alpha = self._alpha
        # initialize matrix Z recursively in IBP manner
        # Debug case
        Z = []
        while len(Z) == 0 or Z.shape[1] == 0:
            Z = np.ones((0, 0), dtype=int)
            for i in range(1, N + 1):
                # sample existing features
                # Z.sum(axis=0)/i: compute the popularity of every dish, computes the probability of sampling that dish
                sample_dish = (np.random.uniform(0, 1, (1, Z.shape[1])) < (Z.sum(axis=0).astype(np.float) / i))
                # sample a value from the poisson distribution, defines the number of new features
                K_new = stats.poisson.rvs((alpha * 1.0 / i))
                # horizontally stack or append the new dishes to current object's observation vector, i.e., the vector Z_{n*}
                sample_dish = np.hstack((sample_dish, np.ones((1, K_new))))
                # append the matrix horizontally and then vertically to the Z matrix
                Z = np.hstack((Z, np.zeros((Z.shape[0], K_new))))
                Z = np.vstack((Z, sample_dish))

        if left_ordered:
            Z = self.leftordered(Z)

        assert(Z.shape[0] == N)
        return Z

    def initialize_Z_debug(self, N=None, alpha=None, KK=9):
        if N is None:
            N = self._N
        if alpha is None:
            alpha = self._alpha
        # initialize matrix Z recursively in IBP manner
        # Debug case
        Z = np.random.randint(0,2, (N, KK))
        #Z = np.ones((1, KK))
        #for i in xrange(2, N + 1):
        #    # sample existing features
        #    # Z.sum(axis=0)/i: compute the popularity of every dish, computes the probability of sampling that dish
        #    sample_dish = (np.random.uniform(0, 1, (1, Z.shape[1])) < (Z.sum(axis=0).astype(np.float) / i))
        #    # sample a value from the poisson distribution, defines the number of new features
        #    K_new = 0
        #    # horizontally stack or append the new dishes to current object's observation vector, i.e., the vector Z_{n*}
        #    sample_dish = np.hstack((sample_dish, np.ones((1, K_new))))
        #    # append the matrix horizontally and then vertically to the Z matrix
        #    Z = np.hstack((Z, np.zeros((Z.shape[0], K_new))))
        #    Z = np.vstack((Z, sample_dish))

        assert(Z.shape[0] == N)
        return Z

    """
    compute the log-likelihood of the Z matrix """
    def log_likelihood_Z(self):
        # compute {K_+} \log{\alpha} - \alpha * H_N, where H_N = \sum_{j=1}^N 1/j
        H_N = np.array([range(self._N)]) + 1.0
        H_N = np.sum(1.0 / H_N)
        log_likelihood = self._K * np.log(self._alpha) - self._alpha * H_N

        # compute the \sum_{h=1}^{2^N-1} \log{K_h!}
        Z_h = np.sum(self._Z, axis=0).astype(np.int)
        Z_h = list(Z_h)
        for k_h in set(Z_h):
            log_likelihood -= math.log(math.factorial(Z_h.count(k_h)))

        # compute the \sum_{k=1}^{K_+} \frac{(N-m_k)! (m_k-1)!}{N!}
        for k in range(self._K):
            m_k = Z_h[k]
            temp_var = 1.0
            if m_k - 1 < self._N - m_k:
                for k_prime in range(self._N - m_k + 1, self._N + 1):
                    if m_k != 1:
                        m_k -= 1

                    temp_var /= k_prime
                    temp_var *= m_k
            else:
                n_m_k = self._N - m_k
                for k_prime in range(m_k, self._N + 1):
                    temp_var /= k_prime
                    temp_var += n_m_k
                    if n_m_k != 1:
                        n_m_k -= 1

            log_likelihood += np.log(temp_var)

        return log_likelihood

    """
    sample alpha from conjugate posterior """
    def sample_alpha(self):
        assert(self._alpha_hyper_parameter != None)
        assert(type(self._alpha_hyper_parameter) == tuple)

        (alpha_hyper_a, alpha_hyper_b) = self._alpha_hyper_parameter

        K_plus = (self._Z.sum(axis=0) > 1).sum()
        posterior_shape = alpha_hyper_a + K_plus
        H_N = np.array([range(self._N)]) + 1.0
        H_N = np.sum(1.0 / H_N)
        posterior_rate = alpha_hyper_b + H_N # best convergence but longer (K get higher)
        #posterior_scale = alpha_hyper_b + self._N # LL lower but faster (K smaller)

        # Posterior in PyIBP ?
        #m = (self._Z != 0).astype(np.int).sum(axis=0)
        #posterior_shape = alpha_hyper_a + m.sum()
        #posterior_scale = float(1) / (alpha_hyper_b + self._N)

        alpha_new = np.random.gamma(posterior_shape, 1/posterior_rate, size=5).mean()
        lgg.debug('hyper sample: alpha: %s' % alpha_new )

        return alpha_new

    """
    sample standard deviation of a multivariant Gaussian distribution
    @param sigma_hyper_parameter: the hyper-parameter of the gamma distribution
    @param matrix: a r*c matrix drawn from a multivariant c-dimensional Gaussian distribution with zero mean and identity c*c covariance matrix """
    @staticmethod
    def sample_sigma(sigma_hyper_parameter, matrix):
        assert(sigma_hyper_parameter != None);
        assert(matrix != None);
        assert(type(sigma_hyper_parameter) == tuple);
        assert(type(matrix) == np.ndarray);

        (sigma_hyper_a, sigma_hyper_b) = sigma_hyper_parameter;
        (row, column) = matrix.shape;

        # compute the posterior_shape = sigma_hyper_a + n/2, where n = self._D * self._K
        posterior_shape = sigma_hyper_a + 0.5 * row * column;
        # compute the posterior_scale = sigma_hyper_b + sum_{k} (A_k - \mu_A)(A_k - \mu_A)^\top/2
        var = 0;
        if row >= column:
            var = np.trace(np.dot(matrix.transpose(), matrix));
        else:
            var = np.trace(np.dot(matrix, matrix.transpose()));

        posterior_scale = 1.0 / (sigma_hyper_b + var * 0.5);
        tau = stats.gamma.rvs(posterior_shape, scale=posterior_scale);
        sigma_a_new = np.sqrt(1.0 / tau);

        return sigma_a_new;

    def leftordered(self, Z=None):
        """ Returns the given matrix in left-ordered-form. """
        if Z is None:
            Z = self._Z
        l = list(Z.T)
        # tuple sorting: sort by first element, then second,etc
        l.sort(key=tuple)
        return np.array(l)[::-1].T

    """
    center the data, i.e., subtract the mean """
    @staticmethod
    def center_data(data):
        (N, D) = data.shape
        data = data - np.tile(data.mean(axis=0), (N, 1))
        return data

    def inititalize_stickbreaking_Z(customers, alpha=10, reducedprop=1.):
        """ Simple implementation of the Indian Buffet Process. Generates a binary matrix with
        customers rows and an expected number of columns of alpha * sum(1,1/2,...,1/customers).
        This implementation uses a stick-breaking construction.
        An additional parameter permits reducing the expected number of times a dish is tried. """
        # max number of dishes is distributed according to Poisson(alpha*sum(1/i))
        _lambda = alpha * np.sum(1. / np.array(list(range(1, customers + 1))))
        alpha /= reducedprop
        # we give it 2 standard deviations as cutoff
        maxdishes = int(_lambda + np.sqrt(_lambda) * 2) + 1
        res = np.zeros((customers, maxdishes), dtype=bool)
        stickprops = np.beta(alpha, 1, maxdishes) # nu_i
        currentstick = 1.
        dishesskipped = 0
        for i, nu in enumerate(stickprops):
            currentstick *= nu
            dishestaken = np.rand(customers) < currentstick * reducedprop
            if np.sum(dishestaken) > 0:
                res[:, i - dishesskipped] = dishestaken
            else:
                dishesskipped += 1
                return res[:, :maxdishes - dishesskipped]

    def plot_matrices(self):
        from plot import ColorMap
        import matplotlib.pyplot as plt
        # define parameter settings
        ps = [(0.1,), (1,), (10,) ]
        N = 30
        generateIBP = self.initialize_Z
        # generate a few matrices, on for each parameter setting
        ms = []
        for p in ps:
                m = generateIBP(N, p[0])
                ms.append(self.leftordered(m))
        # plot the matrices
        for m in ms:
            ColorMap(m, pixelspervalue=6)
        plt.draw()



from collections import defaultdict

import numpy as np
import scipy as sp

#from pymake.frontend.frontend import Object

from .utils import *
from .math import *

import logging
lgg = logging.getLogger('root')

# Optimize Algothim (cython, etc) ?
# frontend -- data copy -- integration ?

class Algo(object):
    def __init__(self, *args, **kwargs):
        pass

    def get_clusters(self, state=dict(), true_order=True):
        """ return cluster of each element in the original order:
            pi: the linear partition of clusters
            labels: the labels of the original orders
            C: number of clusters
        """
        pi = state.get('pi', self.pi)

        C = len(pi)
        # ordered elements
        clusters = np.asarray(sum([ [i] * len(pi[i]) for i in np.arange(C) ], []))

        if true_order is True:
            # Original order
            return clusters[self.labels]
        else:
            return clusters

    def partition(self, state=dict()):
        clusters = self.get_clusters(state)
        K = self.K
        return dict(zip(*[np.arange(K), clusters]))

try:
    import community as pylouvain
except ImportError as e:
    print('Import Error: %s' % e)

class Louvain(Algo):
    # Ugh : structure ?
    def __init__(self, data, iterations=100, **kwargs):
        super(Louvain, self).__init__(**kwargs)
        #nxG do a copy of data ?
        self.data = data
        #self.data = data.copy()

    def search(self):
        g = nxG(self.data)
        self._partition = pylouvain.best_partition(g, resolution=1)
        return list(self._partition.values())

    def partition(self):
        return self._partition

    def _get_clusters(self):
        return list(self._partition.values())

    @staticmethod
    def get_clusters(data, resolution=0.5):
        g = nxG(data)
        _partition = pylouvain.best_partition(g, resolution=resolution)
        _c = list(_partition.values())
        c_i = list(np.unique(_c))
        clusters = [c_i.index(v) for v in _c]
        return clusters


#import warnings
#warnings.filterwarnings('error')

from copy import deepcopy
#from pymake.frontend.frontendnetwork import frontendNetwork

class Annealing(Algo):
#class Annealing(Algo, frontendNetwork):
    """ Find near optimal partitionning of a square matrix (0,1),
        by using a modularity that maximize communities detection.
        This is a kind of Simulated Annealing (SA) that maximize
        the so called energy (instead of minimizing in SA, get crazy ;)
            * maximize inner-clusters energy I
            * minimize inter-cluster energy O
            * Modularity = O - I

        @TODO: Local Optima escape:
            * multinomial instead of argmax ?
            * convex ?

        parameters
        ---------
        data: np.array (K,K)
            A square matrix
        C_init: int
            Number of clusters at start
        iterations: int
            Iterations for boundary search
        grow_rate : int
            Number of clusters to add a each super-iteration
            if 0, just one super-iteration.
    """

    def __init__(self, data, iterations=200, C_init=2, grow_rate=1):
        super(Annealing, self).__init__(data=data)
        self.data = data.copy()
        self.grow_rate = grow_rate
        # Number of initial classes
        self.K = data.shape[0]
        # Keep track of class labels
        self.labels = np.arange(self.K)

        # Number of clusters
        C = C_init

        ### Dichotomy init:
        # Clusters Boundary
        # Store for speed purpose
        self.B = np.linspace(0,self.K, C+1).astype(int)
        # Cluters partitionning
        self.pi = self.get_partitions(self.B)
        # Total Energy
        self.E = self.energy()
        # degree
        self.degree = np.asarray([data[k,:].sum() + data[:,k].sum() - data[k,k] for k in range(self.K)])

        # not used...
        self.iterations = iterations

    def set_state(self, state):
        for k, v in state.items():
            setattr(self, k, v)

    def get_state(self):
        state = {'pi': self.pi,
                 'B': self.B,
                 'data': self.data,
                 'labels': self.labels}
        return state

    def get_C(self):
        # or len(pi)
        return len(self.B)-1

    def get_partitions(self, B):
        """ Clusters Partitionning in a linear label """
        C = len(B) - 1
        return [np.arange(B[i], B[i+1]) for i in range(C)]

    def modularity(self, state=dict()):
        g = self.getG()
        part = self.partition(state)
        modul = pylouvain.modularity(part, g)
        return modul

    #def energy(self, state=dict(), get_params=False): return self.modularity(state)
    def energy(self, state=dict(), get_params=False):
        data = state.get('data', self.data)
        B = state.get('B', self.B)
        pi = state.get('pi', self.get_partitions(B))
        K = self.K
        C = len(pi)

        # Inner Energy
        #blocks = [ np.ix_(*[slice]*2) for slice in pi ]
        diag_blocks = [data[np.ix_(*[slice]*2)] for slice in pi]
        I, L_i = np.asarray(list(zip(* [(a.sum(), float(a.size)) for a in diag_blocks])))

        # Outer Energy
        O = []
        for c in range(C):
            b_low = B[c]
            b_high = B[c+1]
            O.append( data[b_low:b_high, :].sum() + data[:, b_low:b_high].sum() - 2 * I[c])
        L_o = 2 * (K * np.asarray(list(map(len, pi))).astype(float) - L_i)

        if get_params is True:
            return I, L_i, np.asarray(O), L_o

        # Total Energy
        E = 1/float(C) * ((I/L_i).sum() - np.square((O/L_o).sum()))
        #return 1/(1+np.exp(-E))
        #regul =  1 / (np.array(len(pi)).astype(float).sum())**2
        regul = 0
        return E + regul

    def boundary_sample(self, new=0, it=None):
        """ Sample new boundary for clusters
            return new Energy.
        """
        K = self.K
        C = self.get_C() + new

        if it == 0:
            B_new = np.linspace(0,self.K, C+1).astype(int)
        else:
            # Min separation of boundary set to 2.
            B_new = np.sort(np.random.choice(np.arange(2, K-1, 2), C-1, replace=False))
            B_new = np.hstack((0, B_new, K))

        # reload partitioning
        pi_new = self.get_partitions(B_new)

        state = {'B': B_new, 'pi':pi_new}
        return state

    def concentrate_clases(self, state=dict()):
        """ Reorder data by Assignin classes to cluster
            that maximise energy. (descending order).

            Classe will be moving the clusters of choice
            by increasing its size (B updates)
            * Inplace changes.
        """
        data = state.get('data', self.data)
        B = state.get('B', self.B)
        pi = state.get('pi', self.pi)

        K = self.K
        C = len(pi)

        size = np.asarray(list(map(len, pi)))
        clusters = self.get_clusters(state, true_order=False, )

        e_k_in_c = np.zeros((K, C))
        cursor = np.arange(K)
        I, L_i, O, L_o = self.energy(get_params=True)
        for k in np.random.choice(range(K), K, replace=False):
            _k = np.argmax(cursor == k)
            old_c = clusters[k]
            old_i_c = data[_k, B[old_c]:B[old_c+1]].sum() + data[B[old_c]:B[old_c+1], _k].sum() - data[_k,_k]
            degree_k = self.degree[self.labels[k]]
            #E_old = self.energy()
            params_E_tmp = np.zeros((C,4,2))
            for c in np.arange(C):
                #if c == old_c:
                #    #e_k_in_c[k, old_c] = 0
                #    continue
                b_low = B[c]
                b_high = B[c+1]

                ## True update (local optima !)
                #i_c = data[_k, b_low:b_high].sum() + data[b_low:b_high, _k].sum() + data[_k,_k]
                #o_c = degree_k - 2*i_c + data[_k,_k]
                #old_o_c = degree_k - 2*old_i_c + data[_k,_k]
                ##print """ c= %d (true %d), degree_k = %d
                ##        i_c = %d, old_i_c = %d
                ##        o_c = %d, old_o_c = %d
                ##        """ % (c, old_c, degree_k, i_c, old_i_c, o_c, old_o_c)

                #_I   = I[[c  , old_c]] + [i_c                  , - old_i_c]
                #_O   = O[[c  , old_c]] + [o_c                  , - old_o_c]
                #_L_i = L_i[[c, old_c]] + [2*size[c]+1          , - (2*size[old_c] - 1)]
                #_L_o = L_o[[c, old_c]] + [2*(K - 2*size[c] - 1), - 2*(K - 2*size[old_c] + 1)]
                #params_E_tmp[c, 0] = _I
                #params_E_tmp[c, 1] = _L_i
                #params_E_tmp[c, 2] = _O
                #params_E_tmp[c, 3] = _L_o
                ## Delta Energy
                #pt = [c, old_c]
                #e_k_in_c[k, c] = 1/float(C) * ((_I/_L_i).sum() - np.square((_O/_L_o).sum()) - ((I[pt]/L_i[pt]).sum() - np.square((O[pt]/L_o[pt]).sum()) ))

                # Approx update
                sgn = [-1, 1]
                inc = bool(c != old_c)
                i_c = data[_k, b_low:b_high].sum() + data[b_low:b_high, _k].sum() + sgn[inc] *  data[_k,_k]
                o_c = degree_k - i_c
                e_k_in_c[k, c] = i_c / ((size[c] + inc)**2) - o_c / ((K - (size[c] +inc))**2)

            # Keep the clusters with highest energy
            new_c = np.argmax(e_k_in_c[k])
            # new_c == old_c
            if e_k_in_c[k, new_c] == 0: continue
            # update data, size, B and cursor
            new_k = B[new_c]
            shiftpos(cursor, _k, new_k)
            shiftpos(data, _k, new_k, axis=0)
            shiftpos(data, _k, new_k, axis=1)
            size[old_c] -= 1
            size[new_c] += 1
            # @todo: check if cluster disapear...
            #print 'B cluster update %d -> %d (B=%d)' % (old_c, new_c, B[1])
            if old_c < new_c:
                B[1:C][old_c:new_c] -= 1
            else:
                B[1:C][new_c:old_c] += 1
            ## Approx update
            ##I, L_i, O, L_o = self.energy(get_params=True)
            #I[[new_c, old_c]] = params_E_tmp[new_c, 0]
            #L_i[[new_c, old_c]] = params_E_tmp[new_c, 1]
            #O[[new_c, old_c]] = params_E_tmp[new_c, 2]
            #L_o[[new_c, old_c]] = params_E_tmp[new_c, 3]

            #delta_e = (self.energy() - E_old) - e_k_in_c[k, new_c]
            #print 'assert delta_e = 0 --', delta_e
            #print 'I', I[[new_c,old_c]], params_E_tmp[new_c,0]
            #print 'O', O[[new_c,old_c]], params_E_tmp[new_c,2]
            #print 'L_i', L_i[[new_c,old_c]], params_E_tmp[new_c,1]
            #print 'L_o', L_o[[new_c,old_c]], params_E_tmp[new_c,3]
            #if delta_e != 0:
            #    exit()

        #c_hist[::-1].sort() # sort in place
        self.pi = self.get_partitions(B)
        self.labels = self.labels[cursor]

        if self.get_C() != C:
            print( 'heeeere, some clusters are empty!!!!\n need to manage that and eventually remove clusters ?!! %d %d' % (self.get_C(), C))

        #return  1/float(C) * ((I/L_i).sum() - (O/L_o).sum())
        return self.energy()


    def sample_B(self):
        """ Sample Boundary """
        #self.iterations = self.K * self.get_C() / 2
        self.iterations = 100
        self.iterations = int(np.log(self.K) * self.get_C())
        self.iterations = 10

        for it in range(self.iterations): # Broadcast ?!
            print( it, end='')

            state = self.boundary_sample(it=it)
            E_new = self.energy(state)
            #if E_new > self.E :
            if E_new > self.E or self.anneal_transition(E_new, it):
                print (' acceptance %f, %f' %( E_new, self.E))
                self.set_state(state)
                self.E = self.concentrate_clases()
                print ('total Energy %f, B: %s' % (self.E, self.B))
        return

    def search(self):
        """ Sample partionning, until energy reach a maximum
            or reach max iterations
        """
        E_old = self.E
        old_state = self.get_state()
        while (self.get_C() < self.K):
            ### altern Reorder Boundaries - Reorder membership
            self.sample_B()

            #break
            if self.E > E_old and self.grow_rate > 0:
                ### Insert new cluster or keep the current state
                print ('adding new clusters')
                old_state = deepcopy(self.get_state())
                E_old = self.E
                self.set_state(self.boundary_sample(new=self.grow_rate))
            else:
                self.set_state(old_state)
                print ('replacing previsous state (%d cluster)' % self.get_C())
                #self.sample_B()
                break

        return self.get_clusters()

    def anneal_transition(self, E_new, it):
        if it == 0: return True
        T = float(self.iterations - it)
        #T = T / (self.K**2)
        T = T / (self.K**2 / self.get_C())
        delta = self.E - E_new
        p = np.exp(-(delta)/T * np.log(1000))
        #print 'anneal proba %f' %p
        return bool(p > np.random.random())

    def add_cluster(self, n=1):
        """ Add a new clusters in the partitionning """
        pass

    def hi_phi(self):
        """ Return the reordered data with high energy clusters """
        return self.data
    def get_labels(self):
        """ Return the reordered labels"""
        return self.labels

    def stop_criteria(self):
        return False

from scipy.stats import kstest, ks_2samp
from scipy.special import zeta
# Ref: Clauset, Aaron, Cosma Rohilla Shalizi, and Mark EJ Newman. "Power-law distributions in empirical data."
# Debug
# @todo: Estimation of x_min instead of the "max" heuristic.
# @todo: cut-off
def gofit(x, y, model='powerlaw', precision=0.03):
    """ (x, y): the empirical distribution with x the values and y **THE COUNTS** """
    y = y.astype(float)
    #### Power law Goodness of fit
    # Estimate x_min
    y_max = y.max()

    # Reconstruct the data samples
    data = degree_hist_to_list(x, y)
    #x = np.arange(1, y.sum()) ?

    # X_min heuristic /Estim
    index_min = len(y) - np.argmax(y[::-1]) # max from right
    #index_min = np.argmax(y) # max from left
    #x_min = x[index_min]
    index_min = 0
    x_min = x[index_min]
    while x_min == 0:
        index_min += 1
        x_min = x[index_min]

    ### cutoff ?
    x_max = x.max()

    # Estimate \alpha
    N = int(y.sum())
    n_tail = y[index_min:].sum()
    if n_tail < 25 or len(y) < 5:
        # no enough point
        lgg.error('Not enough samples %s' % n_tail)
        return
    #elif n_tail / N < 3/4.0:
    #    # tail not relevant
    #    index_min = len(y) - np.argmax(y[::-1]) # max from left
    #    #index_min = 0 # all the distribution
    #    x_min = x[index_min]
    #    n_tail = y[index_min:].sum()

    alpha = 1 + n_tail * (np.log(data[data>x_min] / (x_min -0.5)).sum())**-1

    ### Build The hypothesis
    if model == 'powerlaw':
        ### Discrete CDF (gives worse p-value)
        cdf = lambda x: 1 - zeta(alpha, x) / zeta(alpha, x_min)
        ### Continious CDF
        #cdf = lambda x:1-(x/x_min)**(-alpha+1)
    else:
        lgg.error('Godfit: Hypothese Unknow %s' % model)
        return

    # Number of synthetic datasets to generate #precision = 0.03
    S = int(0.25 * (precision)**-2)
    pvalue = []

    # Ignore head data
    if False:
        N = n_tail # bad effect
        ks_d = kstest(data[data>=x_min], cdf)
    else:
        ks_d = kstest(data, cdf)

    for s in range(S):
        ### p-value with Kolmogorov-Smirnov, for each synthetic dataset
        # Each synthetic dataset has following size:
        powerlaw_samples_size = np.random.binomial(N, n_tail/N)
        # plus random sample from before the cut
        out_empirical_samples_size = N - powerlaw_samples_size

        # Generate synthetic dataset
        ratio_plaw = 1
        ratio_random = 1
        powerlaw_samples = random_powerlaw(alpha, x_min, powerlaw_samples_size*ratio_plaw)

        if len(data[data<=x_min]) > 0:
            out_samples = np.random.choice((data[data<=x_min]), size=out_empirical_samples_size*ratio_random)
            sync_samples = np.hstack((out_samples, powerlaw_samples))
        else:
            sync_samples = powerlaw_samples

        ### Cutoff ?!
        #powerlaw_samples = powerlaw_samples[powerlaw_samples <= x_max]
        #ratio =  powerlaw_samples_size / len(powerlaw_samples)
        #if ratio > 1:
        #    supplement = random_powerlaw(alpha, x_min, powerlaw_samples_size * (ratio -1))
        #    supplement = supplement[supplement <= x_max]
        #    sync_samples = np.hstack((out_samples, powerlaw_samples, supplement))
        #else:
        #    sync_samples = np.hstack((out_samples, powerlaw_samples))


        #ks_2 = ks_2samp(sync_samples, d)
        ks_s = kstest(sync_samples, cdf)
        pvalue.append(ks_s.statistic >= ks_d.statistic)

    #frontend.DataBase.save(d, 'd.pk')
    #frontend.DataBase.save(sync_samples, 'sc.pk')

    pvalue = sum(pvalue) / len(pvalue)
    estim = {'alpha': alpha, 'x_min':x_min, 'y_max':y_max,
             'n_tail': int(n_tail),'n_head':N - n_tail,
             'pvalue':pvalue}
    print ('KS data: ', ks_d)
    print ('KS synthetic: ', ks_s)
    print (estim)
    estim.update({'sync':sync_samples.astype(int)})
    return estim



######################################
### External
#####################################

try:
    from sklearn.cluster import KMeans
    from sklearn.datasets.samples_generator import make_blobs
except:
    pass
def kmeans(M, K=4):
    km = KMeans( init='k-means++',  n_clusters=K)
    km.fit(M)
    clusters = km.predict(M.astype(float))
    return clusters.astype(int)

from matplotlib import pyplot as plt
def kmeans_plus(X=None, K=4):
    if X is None:
        centers = [[1, 1], [-1, -1], [1, -1]]
        K = len(centers)
        X, labels_true = make_blobs(n_samples=3000, centers=centers, cluster_std=0.7)

    ###################################
    # Compute clustering with Means

    k_means = KMeans(init='k-means++', n_clusters=K, n_init=10)
    k_means.fit(X)
    k_means_labels = k_means.labels_
    k_means_cluster_centers = k_means.cluster_centers_
    k_means_labels_unique = np.unique(k_means_labels)

    ###################################
    # Plot result

    colors = ['#4EACC5', '#FF9C34', '#4E9A06', '#4E9A97']
    plt.figure()
    plt.hold(True)
    for k, col in zip(range(K), colors):
    #for k in range(K):
        my_members = k_means_labels == k
        cluster_center = k_means_cluster_centers[k]
        plt.plot(X[my_members, 0], X[my_members, 1], 'w',
                 markerfacecolor=col, marker='.')
        plt.plot(cluster_center[0], cluster_center[1], 'o',
                 markeredgecolor='k', markersize=6, markerfacecolor=col)
    plt.title('KMeans')
    plt.grid(True)
    plt.show()




import numpy as np
import scipy as sp
import scipy.stats #sp.stats fails if not
from scipy.special import gammaln
from numpy.random import dirichlet, gamma, poisson, binomial, beta
from pymake.util.math import lognormalize, categorical

import logging
lgg = logging.getLogger('root')


# lambda fail to find import if _stirling if not
# visible in the global scope.
import sympy
from sympy.functions.combinatorial.numbers import stirling
try:
    from pymake.util.compute_stirling import load_stirling
    _stirling_mat = load_stirling()
    STIRLING_LOADED = True
except Exception as e:
    print('Stirling matrix cant be loaded:')
    print('error: ', e)
    STIRLING_LOADED = False

class MSampler(object):

    if STIRLING_LOADED:
        stirling_mat = lambda  _, x, y : _stirling_mat[x, y]
    else:
        lgg.error('stirling.npy file not found, using sympy instead (MMSB_CGS model will be 100 time slower !)')
        stirling_mat = lambda  _,x,y : np.asarray([float(sympy.log(stirling(x, i, kind=1)).evalf()) for i in y])

    def __init__(self, zsampler):

        self.zsampler = zsampler
        self.get_log_alpha_beta = zsampler.get_log_alpha_beta
        self.count_k_by_j = zsampler.doc_topic_counts

        # We don't know the preconfiguration of tables !
        self.m = np.ones(self.count_k_by_j.shape, dtype=int)
        self.m_dotk = self.m.sum(axis=0)

    def sample(self):
        self._update_m()

        indices = np.ndenumerate(self.count_k_by_j)

        lgg.debug('Sample m...')
        for ind in indices:
            j, k = ind[0]
            count = ind[1]

            if count > 0:
                # Sample number of tables in j serving dishe k
                params = self.prob_jk(j, k)
                sample = categorical(params) + 1
            else:
                sample = 0

            self.m[j, k] = sample

        self.m_dotk = self.m.sum(0)
        self.purge_empty_tables()

        return self.m

    def _update_m(self):
        # Remove tables associated with purged topics
        for k in sorted(self.zsampler.last_purged_topics, reverse=True):
            self.m = np.delete(self.m, k, axis=1)

        # Passed by reference, but why not...
        self.count_k_by_j = self.zsampler.doc_topic_counts
        K = self.count_k_by_j.shape[1]
        # Add empty table for new fancy topics
        new_k = K - self.m.shape[1]
        if new_k > 0:
            lgg.debug( 'msampler: %d new topics' % (new_k))
            J = self.m.shape[0]
            self.m = np.hstack((self.m, np.zeros((J, new_k), dtype=int)))

    # Removes empty table.
    def purge_empty_tables(self):
        # cant be.
        pass

    def prob_jk(self, j, k):
        # -1 because table of current sample topic jk, is not conditioned on
        njdotk = self.count_k_by_j[j, k]
        if njdotk == 1:
            return np.ones(1)

        possible_ms = np.arange(1, njdotk) # +1-1
        log_alpha_beta_k = self.get_log_alpha_beta(k)
        alpha_beta_k = np.exp(log_alpha_beta_k)

        normalizer = gammaln(alpha_beta_k) - gammaln(alpha_beta_k + njdotk)
        log_stir = self.stirling_mat(njdotk, possible_ms)

        params = normalizer + log_stir + possible_ms*log_alpha_beta_k

        return lognormalize(params)

class BetaSampler(object):

    def __init__(self, gmma, msampler):
        self.gmma = gmma
        self.msampler = msampler

        # Initialize restaurant with just one table.
        self.beta = dirichlet([1, gmma])

    def sample(self):
        lgg.debug( 'Sample Beta...')
        self._update_dirichlet_params()
        self.beta = dirichlet(self.dirichlet_params)

        return self.beta

    def _update_dirichlet_params(self):
        m_dotk_augmented = np.append(self.msampler.m_dotk, self.gmma)
        lgg.debug( 'Beta Dirichlet Prior: %s, alpha0: %.4f ' % (m_dotk_augmented, self.msampler.zsampler.alpha_0))
        self.dirichlet_params = m_dotk_augmented




""" Pygraphviz colorsceme
	ref: http://www.graphviz.org/doc/info/colors.html
"""

#SVG_colors = ["aliceblue", "antiquewhite", "aqua", "aquamarine", "azure","beige", "bisque", "black", "blanchedalmond", "blue", "blueviolet", "brown", "burlywood", "cadetblue", "chartreuse","chocolate", "coral", "cornflowerblue", "cornsilk", "crimson","cyan", "darkblue", "darkcyan", "darkgoldenrod", "darkgray","darkgreen", "darkgrey", "darkkhaki", "darkmagenta", "darkolivegreen","darkorange", "darkorchid", "darkred", "darksalmon", "darkseagreen","darkslateblue", "darkslategray", "darkslategrey", "darkturquoise", "darkviolet","deeppink", "deepskyblue", "dimgray", "dimgrey", "dodgerblue","firebrick", "floralwhite", "forestgreen", "fuchsia", "gainsboro","ghostwhite", "gold", "goldenrod", "gray", "grey", "green", "greenyellow", "honeydew", "hotpink", "indianred","indigo", "ivory", "khaki", "lavender", "lavenderblush","lawngreen", "lemonchiffon", "lightblue", "lightcoral", "lightcyan","lightgoldenrodyellow", "lightgray", "lightgreen", "lightgrey", "lightpink","lightsalmon", "lightseagreen", "lightskyblue", "lightslategray", "lightslategrey","lightsteelblue", "lightyellow", "lime", "limegreen", "linen","magenta", "maroon", "mediumaquamarine", "mediumblue", "mediumorchid","mediumpurple", "mediumseagreen", "mediumslateblue", "mediumspringgreen", "mediumturquoise","mediumvioletred", "midnightblue", "mintcream", "mistyrose", "moccasin","navajowhite", "navy", "oldlace", "olive", "olivedrab","orange", "orangered", "orchid", "palegoldenrod", "palegreen","paleturquoise", "palevioletred", "papayawhip", "peachpuff", "peru", "pink", "plum", "powderblue", "purple", "red", "rosybrown", "royalblue", "saddlebrown", "salmon", "sandybrown","seagreen", "seashell", "sienna", "silver", "skyblue","slateblue", "slategray", "slategrey", "snow", "springgreen","steelblue", "tan", "teal", "thistle", "tomato","turquoise", "violet", "wheat", "white", "whitesmoke","yellow", "yellowgreen"]


X11_colors = ["aliceblue" , "antiquewhite4" , "aquamarine4" , "azure4", "bisque3" , "blue1"  	, "brown"  , "burlywood" , "cadetblue" , "chartreuse", "chocolate" , "coral"	, "cornflowerblue" 	, "cornsilk4" 	, "cyan3" 	 	, "darkgoldenrod3" 	, "darkolivegreen1" , "darkorange1" 	, "darkorchid1" 	, "darkseagreen" 	, "darkslateblue" 	, "darkslategray4" 	, "deeppink1" 	, "deepskyblue1" 	, "dimgrey" 	, "dodgerblue4" 	, "firebrick4" , "gold" 	, "goldenrod" 	, "gray" ]


""" Matplotlib
	ref: http://matplotlib.org/examples/color/named_colors.html
"""
from matplotlib import colors
plt_colors = [(k,v) for k,v in colors.cnames.items()]



import pystan
import numpy as np
import cPickle

# Hyper-parameters
K = 3 # number of communities
alpha = 0.1 * np.ones(K) # Dirichlet prior for membership mixture
block_structure = np.asarray([[0.8, 0.3, 0.2],
                              [0.2, 0.7, 0.3],
                              [0.1, 0.1, 0.6]])

# Generate M testing binary graph
M = 7
N = 10
nodes_community = np.random.randint(0, K, size = N) # values in [0, K-1]
graph_views = []
for i in xrange(M):
    binary_graph = np.diag(np.ones(N, dtype = int))
    for u in xrange(N):
        for v in xrange(u+1, N):
            u_label = nodes_community[u]
            v_label = nodes_community[v]
            edge_prob = block_structure[u_label, v_label]
            if np.random.binomial(1, edge_prob) == 1:
                binary_graph[u, v] = 1
                binary_graph[v, u] = 1

    graph_views += [binary_graph]

graph_views = np.asarray( graph_views )

# Run MCMC
def load_stan_model( model_name ):
    """
    Load stan model from disk,
    if not exist, compile the model from source code
    """
    try:
        stan_model = cPickle.load( open(model_name + ".model", 'rb') )
    except IOError:
        stan_model = pystan.StanModel( file = model_name + ".stan" )
        with open(model_name + ".model", 'wb') as fout:
            cPickle.dump(stan_model, fout)
        pass

    return stan_model



def mmsb_single_view(idx):
    mmsb_model = load_stan_model("mix_membership_blockmodel")
    fit = mmsb_model.sampling( data = {'K': K, 'N': N, 'En': graph_views[idx],
                                       'alpha': alpha, 'Blk': block_structure},
                               chains = 4 )

    # Show the result
    print(fit)
    mu = fit.extract(permuted = True)['mu']
    mean_membership = mu.mean(axis = 0)
    print(mean_membership)
    print("%d out of %d predicted right" % (sum(mean_membership.argmax(axis = 1) == nodes_community), N))


# Multiview MMSB model
mmsb_model = load_stan_model("mmsb_multiview")
fit = mmsb_model.sampling( data = {'K': K, 'N': N, 'En': graph_views,
                                   'alpha': alpha, 'Blk': block_structure} )
print(fit)
mu = fit.extract(permuted = True)['mu']
mean_membership = mu.mean(axis = 0)
print(mean_membership)
print("%d out of %d predicted right" % (sum(mean_membership.argmax(axis = 1) == nodes_community), N))



"""
Simple matrix intensity plot, similar to MATLAB imagesc()

David Andrzejewski (david.andrzej@gmail.com)
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as MT
import matplotlib.cm as CM

def scaledimage(W, pixwidth=1, ax=None, grayscale=True):
    """
    Do intensity plot, similar to MATLAB imagesc()

    W = intensity matrix to visualize
    pixwidth = size of each W element
    ax = matplotlib Axes to draw on
    grayscale = use grayscale color map

    Rely on caller to .show()
    """

    # N = rows, M = column
    (N, M) = W.shape
    # Need to create a new Axes?
    if(ax == None):
        ax = plt.figure().gca()
    # extents = Left Right Bottom Top
    exts = (0, pixwidth * M, 0, pixwidth * N)
    if(grayscale):
        ax.imshow(W,
                  interpolation='nearest',
                  cmap=CM.gray,
                  extent=exts)
    else:
        ax.imshow(W,
                  interpolation='nearest',
                  extent=exts)

    ax.xaxis.set_major_locator(MT.NullLocator())
    ax.yaxis.set_major_locator(MT.NullLocator())
    return ax

if __name__ == '__main__':
    # Define a synthetic test dataset
    testweights = np.array([[0.25, 0.50, 0.25, 0.00],
                            [0.00, 0.50, 0.00, 0.00],
                            [0.00, 0.10, 0.10, 0.00],
                            [0.00, 0.00, 0.25, 0.75]])
    # Display it
    ax = scaledimage(testweights)
    plt.show()



import os
import numpy as np
import scipy as sp
from numpy import ma
from pymake import ExpTensor, GramExp, ExpeFormat, ExpSpace
from pymake.frontend.manager import ModelManager, FrontendManager



import itertools
from pymake.util.algo import gofit, Louvain, Annealing
from pymake.util.math import reorder_mat, sorted_perm, categorical, clusters_hist
from pymake.util.utils import Now, nowDiff, colored
import matplotlib.pyplot as plt
from pymake.plot import plot_degree, degree_hist, adj_to_degree, plot_degree_poly, adjshow, plot_degree_2, random_degree,  draw_graph_circular, draw_graph_spectral, draw_graph_spring, _markers
from pymake.core.format import tabulate

from sklearn.metrics import roc_curve, auc, precision_recall_curve



USAGE = '''\
----------------
Generate data  --or find the answers :
----------------
 |
 |   methods
 |   ------
 |   burstiness : global + local + feature burstiness.
 |   homo       : homophily based analysis.
 |
>> Examples
    parallel ./generate.py -k {}  ::: $(echo 5 10 15 20)
    generate --alpha 1 --gmma 1 -n 1000 --seed
'''

class GenNetwork(ExpeFormat):

    _default_expe = dict(
        block_plot = False,
        _write  = False,
        _do            = ['burstiness', 'global'], # default
        _mode         = 'generative',
        gen_size      = 1000,
        epoch         = 10 , # Gen,eration epoch
        limit_class   = 15, # Ignored ?
    )

    def _preprocess(self):
        expe = self.expe

        frontend = FrontendManager.load(expe)
        if frontend:
            self._N = frontend.getN()
            expe.symmetric = frontend.is_symmetric()
        else:
            self._N = expe.N
            expe.symmetric = True


        if expe._mode == 'predictive':
            ### Generate data from a fitted model
            model = ModelManager.from_expe(expe, load=True)

            try:
                # this try due to mthod modification entry in init not in picke object..
                expe.hyperparams = model.get_hyper()
            except Exception as e:
                self.log.warning('loading hyperparam error: %s' % e)
                if model is not None:
                    model._mean_w = 0
                    expe.hyperparams = 0

        elif expe._mode == 'generative':
            ### Generate data from a un-fitted model

            expe.alpha = 1
            expe.gmma = 1/2
            expe.delta = [0.5,0.5]

            if 'ilfm' in expe.model:
                keys_hyper = ('alpha','delta')
                hyper = (expe.alpha, expe.delta)
            else:
                keys_hyper = ('alpha','gmma','delta')
                hyper = (expe.alpha, expe.gmma, expe.delta)
            expe.hyperparams = dict(zip(keys_hyper, hyper))
            expe.hyper = 'fix' # dummy
            model = ModelManager.from_expe(expe, load=False)
            #model.update_hyper(hyper)

            ## Obsolete !
            #if 'ilfm' in expe.model:
            #    title = 'N=%s, K=%s alpha=%s, lambda:%s'% ( N, expe.K, expe.alpha, expe.delta)
            #elif 'immsb' in expe.model:
            #    title = 'N=%s, K=%s alpha=%s, gamma=%s, lambda:%s'% (N, expe.K, expe.alpha, expe.gmma, expe.delta)
            #elif 'mmsb' in expe.model:
            #    title = 'N=%s, K=%s alpha=%s, lambda:%s'% ( N, expe.K, expe.alpha, expe.delta)
            #else:
            #    raise NotImplementedError

            #expe.title = title

        else:
            raise NotImplementedError('What generation context ? predictive/generative..')

        self.log.info('=== GenNetworks === ')
        self.log.info('Mode: %s' % expe._mode)
        self.log.info('===')
        self.log.info('hyper: %s' % (str(expe.hyperparams)))

        self.frontend = frontend
        self.model = model

        if model is None:
            raise FileNotFoundError('No model for Expe at :  %s' % self.output_path)


        if expe._do[0] in ('burstiness','pvalue','homo', 'homo_mustach'):
            self._generate()


    def _generate(self):
        ''' Generate data. '''
        expe = self.expe
        model = self.model

        Y = []
        Theta = []
        Phi = []
        now = Now()
        for i in range(expe.epoch):
            y, theta, phi = model.generate(mode=expe._mode,
                                           N=self._N, K=expe.K,
                                           symmetric=expe.symmetric,
                                          )
            Y.append(y)
            Theta.append(theta)
            Phi.append(phi)
        self.log.info('Data Generation : %s second' % nowDiff(now))

        #if expe.symmetric:
        #    for y in Y:
        #        frontend.symmetrize(y)

        self._Y = Y
        self._Theta = Theta
        self._Phi = Phi


    def init_fit_tables(self, _type, Y=[]):
        expe = self.expe
        if not hasattr(self.gramexp, 'tables'):
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            models = self.gramexp.get_set('model')
            Meas = [ 'pvalue', 'alpha', 'x_min', 'n_tail']
            tables = {}
            for m in models:
                if _type == 'local':
                    table_shape = (len(corpuses), len(Meas),2* expe.K**2)
                    table = ma.array(np.empty(table_shape), mask=True)
                elif _type == 'global':
                    table = np.empty((len(corpuses), len(Meas), len(Y)))
                tables[m] = table
            self.gramexp.Meas = Meas
            self.gramexp.tables = tables
            Table = tables[expe.model]
        else:
            Table = self.gramexp.tables[expe.model]
            Meas = self.gramexp.Meas

        return Table, Meas

    def init_roc_tables(self):
        expe = self.expe
        if not hasattr(self.gramexp, 'tables'):
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            if not 'testset_ratio' in self.pt:
                Meas = ['20']
            else:
                Meas = self.gramexp.exp_tensor['testset_ratio']
            tables = ma.array(np.empty((len(corpuses), len(Meas),len(self.gramexp.getK('_repeat')), 2)), mask=True)
            self.gramexp.Meas = Meas
            self.gramexp.tables = tables
        else:
            tables = self.gramexp.tables
            Meas = self.gramexp.Meas

        return tables, Meas

    @ExpeFormat.raw_plot
    def burstiness(self, _type='all'):
        '''Zipf Analysis
           (global burstiness) + local burstiness + feature burstiness

           Parameters
           ----------
           _type : str
            type of burstiness to compute in ('global', 'local', 'feature', 'all')
        '''
        if self.model is None: return
        expe = self.expe
        figs = []

        Y = self._Y
        N = Y[0].shape[0]
        model = self.model

        if _type in ('global', 'all'):
            # Global burstiness
            d, dc, yerr = random_degree(Y)
            fig = plt.figure()
            title = 'global | %s, %s' % (self.specname(expe.get('corpus')), self.specname(expe.model))
            plot_degree_2((d,dc,yerr), logscale=True, title=title)

            figs.append(plt.gcf())

        if _type in  ('local', 'all'):
            # Local burstiness
            print ('Computing Local Preferential attachment')
            a, b = model.get_params()
            N,K = a.shape
            print('theta shape: %s'%(str((N,K))))
            now = Now()
            if 'mmsb' in expe.model:
                ### Z assignement method #
                ZZ = []
                for _i, _ in enumerate(Y):
                #for _ in Y: # Do not reflect real local degree !

                    theta = self._Theta[_i]
                    phi = self._Phi[_i]
                    Z = np.empty((2,N,N))
                    order = np.arange(N**2).reshape((N,N))
                    if expe.symmetric:
                        triu = np.triu_indices(N)
                        order = order[triu]
                    else:
                        order = order.flatten()
                    order = zip(*np.unravel_index(order, (N,N)))

                    for i,j in order:
                        Z[0, i,j] = categorical(theta[i])
                        Z[1, i,j] = categorical(theta[j])
                    Z[0] = np.triu(Z[0]) + np.triu(Z[0], 1).T
                    Z[1] = np.triu(Z[1]) + np.triu(Z[1], 1).T
                    ZZ.append( Z )
                self.log.info('Z formation %s second' % nowDiff(now))

            clustering = 'modularity'
            comm = model.communities_analysis(data=Y[0], clustering=clustering)
            print('clustering method: %s, active clusters ratio: %f' % (
                clustering,
                len(comm['block_hist']>0)/K))

            local_degree_c = {}
            ### Iterate over all classes couple
            if expe.symmetric:
                #k_perm = np.unique( map(list, map(set, itertools.product(np.unique(clusters) , repeat=2))))
                k_perm =  np.unique(list(map(list, map(list, map(set, itertools.product(range(K) , repeat=2))))))
            else:
                #k_perm = itertools.product(np.unique(clusters) , repeat=2)
                k_perm = itertools.product(range(K) , repeat=2)

            fig = plt.figure()
            for i, c in enumerate(k_perm):
                if isinstance(c,(np.int64, np.float64)):
                    k = l = c
                elif len(c) == 2:
                    # Stochastic Equivalence (outer class)
                    k, l = c
                else:
                    # Comunnities (inner class)
                    k = l = c.pop()
                #if i > expe.limit_class:
                #   break
                if k != l:
                    continue

                degree_c = []
                YY = []
                if 'mmsb' in expe.model:
                    for y, z in zip(Y, ZZ): # take the len of ZZ if < Y
                        y_c = np.zeros(y.shape)
                        phi_c = np.zeros(y.shape)
                        # UNDIRECTED !
                        phi_c[(z[0] == k) & (z[1] == l)] = 1
                        y_c = y * phi_c
                        #degree_c += adj_to_degree(y_c).values()
                        #yerr= None
                        YY.append(y_c)
                elif 'ilfm' in  expe.model: # or Corpus !
                    for _i , y in enumerate(Y):
                        theta = self._Theta[_i]
                        if theta.shape[1] <= max(k,l):
                            print('warning: not all block converted.')
                            continue
                        YY.append((y * np.outer(theta[:,k], theta[:,l] )).astype(int))

                d, dc, yerr = random_degree(YY)
                if len(d) == 0: continue
                title = 'local | %s, %s' % (self.specname(expe.get('corpus')), self.specname(expe.model))
                plot_degree_2((d,dc,yerr), logscale=True, colors=True, line=True,
                             title=title)
            figs.append(plt.gcf())

        # Blockmodel Analysis
        #if _type in  ('feature', 'all'):
        #    plt.figure()
        #    if 'mmsb' in expe.model:
        #        # Feature burstiness
        #        hist, label = clusters_hist(comm['clusters'])
        #        bins = len(hist)
        #        plt.bar(range(bins), hist)
        #        plt.xticks(np.arange(bins)+0.5, label)
        #        plt.xlabel('Class labels')
        #        plt.title('Blocks Size (max assignement)')
        #    elif 'ilfm' in expe.model:
        #        # Feature burstiness
        #        hist, label = sorted_perm(comm['block_hist'], reverse=True)
        #        bins = len(hist)
        #        plt.bar(range(bins), hist)
        #        plt.xticks(np.arange(bins)+0.5, label)
        #        plt.xlabel('Class labels')
        #        plt.title('Blocks Size (max assignement)')

        #    figs.append(plt.gcf())

        if expe._write:
            if expe._mode == 'predictive':
                base = '%s_%s' % (self.specname(expe.corpus), self.specname(expe.model))
            else:
                base = '%s_%s' % ('MG', self.specname(expe.model))
            self.write_frames(figs, base=base)
            return



    # @redondancy with burstiness !
    # in @ExpFormat.table
    def pvalue(self, _type='global'):
        """ similar to zipf but compute pvalue and print table

            Parameters
            ==========
            _type: str in [global, local, feature]
        """
        if self.model is None: return
        expe = self.expe
        figs = []

        Y = self._Y
        N = Y[0].shape[0]
        model = self.model

        Table, Meas = self.init_fit_tables(_type, Y)

        self.log.info('using `%s\' burstiness' % _type)

        if _type  == 'global':
            ### Global degree
            for it_dat, data in enumerate(Y):
                d, dc = degree_hist(adj_to_degree(data), filter_zeros=True)
                gof = gofit(d, dc)
                if not gof:
                    continue

                for i, v in enumerate(Meas):
                    Table[self.corpus_pos, i, it_dat] = gof[v]

        elif _type  == 'local':
            ### Z assignement method
            a, b = model.get_params()
            N,K = a.shape
            print('theta shape: %s'%(str((N,K))))
            now = Now()
            if 'mmsb' in expe.model:
                ZZ = []
                for _i, _ in enumerate(Y):
                #for _ in Y: # Do not reflect real local degree !
                    theta = self._Theta[_i]
                    phi = self._Phi[_i]
                    Z = np.empty((2,N,N))
                    order = np.arange(N**2).reshape((N,N))
                    if expe.symmetric:
                        triu = np.triu_indices(N)
                        order = order[triu]
                    else:
                        order = order.flatten()
                    order = zip(*np.unravel_index(order, (N,N)))

                    for i,j in order:
                        Z[0, i,j] = categorical(theta[i])
                        Z[1, i,j] = categorical(theta[j])
                    Z[0] = np.triu(Z[0]) + np.triu(Z[0], 1).T
                    Z[1] = np.triu(Z[1]) + np.triu(Z[1], 1).T
                    ZZ.append( Z )
                self.log.info('Z formation %s second', nowDiff(now))

            clustering = 'modularity'
            comm = model.communities_analysis(data=Y[0], clustering=clustering)
            print('clustering method: %s, active clusters ratio: %f' % (
                clustering,
                len(comm['block_hist']>0)/K))

            local_degree_c = {}
            ### Iterate over all classes couple
            if expe.symmetric:
                #k_perm = np.unique( map(list, map(set, itertools.product(np.unique(clusters) , repeat=2))))
                k_perm =  np.unique(list(map(list, map(list, map(set, itertools.product(range(K) , repeat=2))))))
            else:
                #k_perm = itertools.product(np.unique(clusters) , repeat=2)
                k_perm = itertools.product(range(K) , repeat=2)

            for it_k, c in enumerate(k_perm):
                if isinstance(c,(np.int64, np.float64)):
                    k = l = c
                elif len(c) == 2:
                    # Stochastic Equivalence (extra class bind
                    k, l = c
                    #continue
                else:
                    # Comunnities (intra class bind)
                    k = l = c.pop()
                #if i > expe.limit_class:
                #   break
                if k != l:
                    continue

                degree_c = []
                YY = []
                if 'mmsb' in expe.model:
                    for y, z in zip(Y, ZZ): # take the len of ZZ if < Y
                        y_c = y.copy()
                        phi_c = np.zeros(y.shape)
                        # UNDIRECTED !
                        phi_c[(z[0] == k) & (z[1] == l)] = 1 #; phi_c[(z[0] == l) & (z[1] == k)] = 1
                        y_c[phi_c != 1] = 0
                        #degree_c += adj_to_degree(y_c).values()
                        #yerr= None
                        YY.append(y_c)
                elif 'ilfm' in expe.model:
                    for _i , y in enumerate(Y):
                        theta = self._Theta[_i]
                        YY.append((y * np.outer(theta[:,k], theta[:,l])).astype(int))

                d, dc, yerr = random_degree(YY)
                if len(d) == 0: continue
                gof =  gofit(d, dc)
                if not gof:
                    continue

                for i, v in enumerate(Meas):
                    Table[self.corpus_pos, i, it_k] = gof[v]

        elif _type == 'feature':
            raise NotImplementedError

        if self._it == self.expe_size -1:
            for _model, table in self.gramexp.tables.items():

                # Mean and standard deviation
                table_mean = np.char.array(np.around(table.mean(2), decimals=3)).astype("|S20")
                table_std = np.char.array(np.around(table.std(2), decimals=3)).astype("|S20")
                table = table_mean + b' $\pm$ ' + table_std

                # Table formatting
                corpuses = self.specname(self.gramexp.get_set('corpus'))
                table = np.column_stack((self.specname(corpuses), table))
                tablefmt = 'simple'
                table = tabulate(table, headers=['__'+_model.upper()+'__']+Meas, tablefmt=tablefmt, floatfmt='.3f')
                print()
                print(table)
                if expe._write:
                    if expe._mode == 'predictive':
                        base = '%s_%s_%s' % (self.specname(expe.corpus), self.specname(_model), _type)
                    else:
                        base = '%s_%s_%s' % ('MG', self.specname(_model), _type)
                    self.write_frames(table, base=base, ext='md')

    @ExpeFormat.raw_plot
    def draw(self):
        #if expe._mode == 'predictive':
        #    model = self.frontend
        #    y = model.data
        #    # move this in draw data
        #elif expe._mode == 'generative':
        #    model = self.model
        #    y = model.generate(**expe)
        model = self.model
        y = model.generate(**expe)

        #clustering = 'modularity'
        #print('@heeere, push commmunities annalysis outside static method of frontendnetwork')
        #comm = model.communities_analysis(y, clustering=clustering)
        #clusters = comm['clusters']
        #draw_graph_spring(y, clusters)
        #draw_graph_spectral(y, clusters)
        #draw_graph_circular(y, clusters)

        description = '/'.join((self.expe._refdir, os.path.basename(self.output_path)))
        adjshow(y, title=description)
        #adjblocks(y, clusters=comm['clusters'], title='Blockmodels of Adjacency matrix')
        #adjshow(reorder_mat(y, comm['clusters']), 'test reordering')
        #draw_blocks(comm)


    def homo(self, _type='pearson', _sim='latent'):
        """ Hmophily test -- table output
            Parameters
            ==========
            _type: similarity type in (contengency, pearson)
            _sim: similarity metric in (natural, latent)
        """
        if self.model is None: return
        expe = self.expe
        figs = []

        Y = self._Y
        N = Y[0].shape[0]
        model = self.model

        self.log.info('using `%s\' type' % _type)

        if not hasattr(self.gramexp, 'tables'):
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            models = self.gramexp.get_set('model')
            tables = {}
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            for m in models:
                if _type == 'pearson':
                    Meas = [ 'pearson coeff', '2-tailed pvalue' ]
                    table = np.empty((len(corpuses), len(Meas), len(Y)))
                elif _type == 'contingency':
                    Meas = [ 'natural', 'latent', 'natural', 'latent' ]
                    table = np.empty((2*len(corpuses), len(Meas), len(Y)))
                tables[m] = table

            self.gramexp.Meas = Meas
            self.gramexp.tables = tables
            table = tables[expe.model]
        else:
            table = self.gramexp.tables[expe.model]
            Meas = self.gramexp.Meas

        if _type == 'pearson':
            self.log.info('using `%s\' similarity' % _sim)
            # No variance for link expecation !!!
            Y = [Y[0]]

            ### Global degree
            d, dc, yerr = random_degree(Y)
            sim = model.similarity_matrix(sim=_sim)
            #plot(sim, title='Similarity', sort=True)
            #plot_degree(sim)
            for it_dat, data in enumerate(Y):
                #homo_object = data
                homo_object = model.likelihood()
                table[self.corpus_pos, :,  it_dat] = sp.stats.pearsonr(homo_object.flatten(), sim.flatten())

        elif _type == 'contingency':

            ### Global degree
            d, dc, yerr = random_degree(Y)
            sim_nat = model.similarity_matrix(sim='natural')
            sim_lat = model.similarity_matrix(sim='latent')
            step_tab = len(self.specname(self.gramexp.get_set('corpus')))
            for it_dat, data in enumerate(Y):

                #homo_object = data
                homo_object = model.likelihood()

                table[self.corpus_pos, 0,  it_dat] = sim_nat[data == 1].mean()
                table[self.corpus_pos, 1,  it_dat] = sim_lat[data == 1].mean()
                table[self.corpus_pos, 2,  it_dat] = sim_nat[data == 1].var()
                table[self.corpus_pos, 3,  it_dat] = sim_lat[data == 1].var()
                table[self.corpus_pos+step_tab, 0,  it_dat] = sim_nat[data == 0].mean()
                table[self.corpus_pos+step_tab, 1,  it_dat] = sim_lat[data == 0].mean()
                table[self.corpus_pos+step_tab, 2,  it_dat] = sim_nat[data == 0].var()
                table[self.corpus_pos+step_tab, 3,  it_dat] = sim_lat[data == 0].var()

        if self._it == self.expe_size -1:
            for _model, table in self.gramexp.tables.items():
                # Function in (utils. ?)
                # Mean and standard deviation
                table_mean = np.char.array(np.around(table.mean(2), decimals=3)).astype("|S20")
                table_std = np.char.array(np.around(table.std(2), decimals=3)).astype("|S20")
                table = table_mean + b' $\pm$ ' + table_std

                # Table formatting
                corpuses = self.specname(self.gramexp.get_set('corpus'))
                try:
                    table = np.column_stack((corpuses, table))
                except:
                    table = np.column_stack((corpuses*2, table))
                tablefmt = 'simple' # 'latex'
                table = tabulate(table, headers=['__'+_model.upper()+'__']+Meas, tablefmt=tablefmt, floatfmt='.3f')
                print()
                print(table)
                if expe._write:
                    base = '%s_homo_%s' % (self.specname(_model), _type)
                    self.write_frames(table, base=base, ext='md')

    @ExpeFormat.raw_plot('model')
    def homo_mustach(self, frame):
        """ Hmophily mustach
        """
        if self.model is None: return
        expe = self.expe
        figs = []

        Y = self._Y
        N = Y[0].shape[0]
        model = self.model

        if not hasattr(self.gramexp, 'tables'):
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            models = self.gramexp.get_set('model')
            tables = {}
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            for m in models:
                sim = [ 'natural', 'latent']
                Meas = ['links', 'non-links']
                table = {'natural': {'links':[], 'non-links':[]},'latent': {'links':[], 'non-links':[]} }
                tables[m] = table

            self.gramexp.Meas = Meas
            self.gramexp.tables = tables
            table = tables[expe.model]
        else:
            table = self.gramexp.tables[expe.model]
            Meas = self.gramexp.Meas


        ### Global degree
        d, dc, yerr = random_degree(Y)
        sim_nat = model.similarity_matrix(sim='natural')
        sim_lat = model.similarity_matrix(sim='latent')
        step_tab = len(self.specname(self.gramexp.get_set('corpus')))

        if not hasattr(self.gramexp._figs[expe.model], 'damax'):
            damax = -np.inf
        else:
            damax = self.gramexp._figs[expe.model].damax
        self.gramexp._figs[expe.model].damax = max(sim_nat.max(), sim_lat.max(), damax)
        for it_dat, data in enumerate(Y):

            #homo_object = data
            #homo_object = model.likelihood()

            table['natural']['links'].extend(sim_nat[data == 1].tolist())
            table['natural']['non-links'].extend(sim_nat[data == 0].tolist())
            table['latent']['links'].extend(sim_lat[data == 1].tolist())
            table['latent']['non-links'].extend(sim_lat[data == 0].tolist())


        if self._it == self.expe_size -1:
            for _model, table in self.gramexp.tables.items():
                ax = self.gramexp._figs[_model].fig.gca()

                bp = ax.boxplot([table['natural']['links']    ], widths=0.5,  positions = [1], whis='range')
                bp = ax.boxplot([table['natural']['non-links']], widths=0.5,  positions = [2], whis='range')
                bp = ax.boxplot([table['latent']['links']     ], widths=0.5,  positions = [4], whis='range')
                bp = ax.boxplot([table['latent']['non-links'] ], widths=0.5,  positions = [5], whis='range')

                ax.set_ylabel('Similarity')
                ax.set_xticks([1.5,4.5])
                ax.set_xticklabels(('natural', 'latent'), rotation=0)
                ax.set_xlim(0,6)

                nbox = 4
                top = self.gramexp._figs[_model].damax
                pos = [1,2,4,5]
                upperLabels = ['linked','    non-linked']*2
                #weights = ['light', 'ultralight']
                weights = ['normal', 'normal']
                for tick  in range(nbox):
                    ax.text(pos[tick], top+top*0.015 , upperLabels[tick],
                             horizontalalignment='center', weight=weights[tick%2])

                print(_model)
                t1 = sp.stats.ttest_ind(table['natural']['links'], table['natural']['non-links'])
                t2 = sp.stats.ttest_ind(table['latent']['links'], table['latent']['non-links'])
                print(t1)
                print(t2)

    @ExpeFormat.raw_plot('corpus', 'testset_ratio')
    def roc(self, frame,  _type='testset', _ratio=100):
        ''' AUC/ROC test report '''
        from sklearn.metrics import roc_curve, auc, precision_recall_curve
        expe = self.expe
        model = self.model
        data = self.frontend.data
        _ratio = int(_ratio)
        _predictall = (_ratio >= 100) or (_ratio < 0)
        if not hasattr(expe, 'testset_ratio'):
            setattr(expe, 'testset_ratio', 20)

        ax = frame.ax()

        if _type == 'testset':
            y_true, probas = model.mask_probas(data)
            if not _predictall:
                # take 20% of the size of the training set
                n_d =int( _ratio/100 * data.size * (1 - expe.testset_ratio/100) / (1 - _ratio/100))
                y_true = y_true[:n_d]
                probas = probas[:n_d]
            else:

                # Log Params statistics
                theta, phi = model.get_params()
                print('--- Params stats')
                print('Theta: shape: %s' %(str(theta.shape)))
                print('Theta: max: %s | min: %.4f | mean: %.4f | std: %.4f  ' % (theta.max(), theta.min(), theta.mean(), theta.std()))
                try:
                    print('Phi: shape: %s' %(str(phi.shape)))
                    print('Phi: max: %.4f | min: %.4f | mean: %.4f | std: %.4f  ' % (phi.max(), phi.min(), phi.mean(), phi.std()))
                except:
                    pass
                print('prediction:  links(1): %d | non-links(0): %d' % (y_true.sum(), (y_true==0).sum()))
                print('Prediction: probas stat: mean: %.4f | std: %.4f' % (probas.mean(), probas.std()))
                print('---')

        elif _type == 'learnset':
            n = int(data.size * _ratio)
            mask_index = np.unravel_index(np.random.permutation(data.size)[:n], data.shape)
            y_true = data[mask_index]
            probas = model.likelihood()[mask_index]

        try:
            fpr, tpr, thresholds = roc_curve(y_true, probas)
        except Exception as e:
            print(e)
            self.log.error('cant format expe : %s' % (self.output_path))
            return

        roc_auc = auc(fpr, tpr)
        description = '/'.join((expe._refdir, os.path.basename(self.output_path)))
        #description = self.specname(expe.model)
        ax.plot(fpr, tpr, label='ROC %s (area = %0.2f)' % (description, roc_auc), ls=frame.linestyle.next())
        plt.legend(loc='upper right',prop={'size':5})
        self.noplot = True

        #precision, recall, thresholds = precision_recall_curve( y_true, probas)
        #plt.plot(precision, recall, label='PR curve; %s' % (expe.model ))

        if self._it == self.expe_size -1:
            for c, f in self.gramexp._figs.items():
                ax = f.fig.gca()
                ax.plot([0, 1], [0, 1], linestyle='--', color='k', label='Luck')
                ax.legend(loc="lower right", prop={'size':5})

    def roc_evolution(self, _type='testset', _type2='max', _ratio=20, _type3='errorbar'):
        ''' AUC difference between two models against testset_ratio
            * _type : learnset/testset
            * _type2 : max/min/mean
            * _ratio : ration of the traning set to predict. If 100 _predictall will be true

        '''
        expe = self.expe
        model = self.model
        data = self.frontend.data
        _ratio = int(_ratio)
        _predictall = (_ratio >= 100) or (_ratio < 0)
        if not hasattr(expe, 'testset_ratio'):
            setattr(expe, 'testset_ratio', 20)
            self.testset_ratio_pos = 0
        else:
            self.testset_ratio_pos = self.pt['testset_ratio']

        table, Meas = self.init_roc_tables()

        #mask = model.get_mask()
        if _type == 'testset':
            y_true, probas = model.mask_probas(data)
            if not _predictall:
                # take 20% of the size of the training set
                n_d =int( _ratio/100 * data.size * (1 - expe.testset_ratio/100) / (1 - _ratio/100))
                y_true = y_true[:n_d]
                probas = probas[:n_d]
            else:
                pass

        elif _type == 'learnset':
            n = int(data.size * _ratio)
            mask_index = np.unravel_index(np.random.permutation(data.size)[:n], data.shape)
            y_true = data[mask_index]
            probas = model.likelihood()[mask_index]

        # Just the ONE:1
        #idx_1 = (y_true == 1)
        #idx_0 = (y_true == 0)
        #size_1 = idx_1.sum()
        #y_true = np.hstack((y_true[idx_1], y_true[idx_0][:size_1]))
        #probas = np.hstack((probas[idx_1], probas[idx_0][:size_1]))

        fpr, tpr, thresholds = roc_curve(y_true, probas)
        roc_auc = auc(fpr, tpr)

        table[self.corpus_pos, self.testset_ratio_pos, self.pt['_repeat'], self.model_pos] = roc_auc

        #precision, recall, thresholds = precision_recall_curve( y_true, probas)
        #plt.plot(precision, recall, label='PR curve; %s' % (expe.model ))

        if self._it == self.expe_size -1:

            # Reduce each repetitions
            take_type = getattr(np, _type2)
            t = ma.array(np.empty(table[:,:,0,:].shape), mask=True)
            t[:,:,0] = take_type(table[:, :, :, 0], -1)
            t[:,:,1] = take_type(table[:, :, :, 1], -1)
            table_mean = t.copy()
            t[:,:,0] = table[:, :, :, 0].std(-1)
            t[:,:,1] = table[:, :, :, 1].std(-1)
            table_std = t

            # Measure is comparaison of two AUC.
            id_mmsb = [i for i, s in enumerate(self.gramexp.exp_tensor['model']) if s.endswith('mmsb_cgs')][0]
            id_ibp = 1 if id_mmsb == 0 else 0
            table_mean = table_mean[:,:, id_mmsb] - table_mean[:,:, id_ibp]
            table_std = table_std[:,:, id_mmsb] + table_std[:,:, id_ibp]

            if _type2 != 'mean':
                table_std = [None] * len(table_std)

            fig = plt.figure()
            corpuses = self.specname(self.gramexp.get_set('corpus'))
            for i in range(len(corpuses)):
                if _type3 == 'errorbar':
                    plt.errorbar(list(map(int, Meas)), table_mean[i], yerr=table_std[i],
                                 fmt=_markers.next(),
                                 label=corpuses[i])
                elif _type3 == 'boxplot':
                    bplot = table[i,:,:,0] - table[i,:,:,1]
                    plt.boxplot(bplot.T, labels=corpuses[i])
                    fig.gca().set_xticklabels(Meas)

            plt.errorbar(Meas,[0]*len(Meas), linestyle='--', color='k')
            plt.legend(loc='lower left',prop={'size':7})

            # Table formatting
            #table = table_mean + b' $\pm$ ' + table_std
            table = table_mean

            corpuses = self.specname(self.gramexp.get_set('corpus'))
            table = np.column_stack((self.specname(corpuses), table))
            tablefmt = 'simple'
            headers = ['']+Meas
            table = tabulate(table, headers=headers, tablefmt=tablefmt, floatfmt='.3f')
            print()
            print(table)
            if expe._write:
                base = '%s_%s_%s' % ( _type, _type2, _ratio)
                figs = {'roc_evolution': ExpSpace({'fig':fig, 'table':table, 'base':base})}
                self.write_frames(figs)

    @ExpeFormat.raw_plot
    def clustering(self):
        algo = 'Louvain'
        algo = 'Annealing'
        data = self.frontend.data

        mat = data
        #mat = phi

        alg = getattr(A, algo)(mat)
        clusters = alg.search()

        mat = draw_boundary(alg.hi_phi(), alg.B)
        #mat = draw_boundary(mat, clusters)

        adjshow(mat, algo)
        plt.colorbar()





if __name__ == '__main__':

    GramExp.generate(usage=USAGE).pymake(GenNetwork)




from time import time
import sys
import numpy as np
import scipy as sp
from numpy import ma
import scipy.stats

from pymake.util.math import lognormalize, categorical, sorted_perm, adj_to_degree, gem
from ml.model.modelbase import SVB

#import warnings
#warnings.filterwarnings('error')
#warnings.catch_warnings()
##np.seterr(all='print')


class iwmmsb_scvb2(SVB):

    _purge = ['_kernel', '_lut_nbinom']

    def _init_params(self, frontend):
        self.frontend = frontend

        # Save the testdata
        data_test = np.transpose(self.frontend.data_test.nonzero())
        frontend.reverse_filter()
        weights = []
        for i,j in data_test:
            weights.append(frontend.weight(i,j))
        frontend.reverse_filter()
        self.data_test = np.hstack((data_test, np.array(weights)[None].T))

        # Data statistics
        _len = {}
        _len['K'] = self.expe.get('K')
        _len['N'] = frontend.num_nodes()
        _len['nnz'] = frontend.num_nnz()
        #_len['nnz_t'] = frontend.num_nnz_t()
        _len['dims'] = frontend.num_neighbors()
        _len['nnz_ones'] = frontend.num_edges()
        _len['nnzsum'] = frontend.num_nnzsum()
        self._len = _len

        self._K = self._len['K']
        self._is_symmetric = frontend.is_symmetric()
        self.limit_elbo_diff = 1e-3

        self._init_gradient()

        # Hyperparams
        self.hyper_theta = np.asarray([1.0 / (i + np.sqrt(self._len['K'])) for i in range(self._len['K'])])
        self.hyper_theta /= self.hyper_theta.sum()

        self.hyper_phi = np.asarray(self.expe['delta'])
        hyper_phi = self.expe['delta']
        if hyper_phi == 'auto':
            self.hyper_phi = np.array([10,0.5])
            self._hyper_phi = 'auto'
        elif len(hyper_phi) == 2:
            self.hyper_phi = np.asarray(hyper_phi)
            self._hyper_phi = 'fix'
        else:
            raise ValueError('hyper parmeter hyper_phi dont understood: %s' % hyper_phi)

        self._random_ss_init()

    def _init_gradient(self):
        self._timestep_a = 0
        self._timestep_b = 0
        self._timestep_c = 0
        self._chi_a = self.expe.get('chi_a', 5)
        self._tau_a = self.expe.get('tau_a', 10)
        self._kappa_a = self.expe.get('kappa_a', 0.9)
        self._chi_b = self.expe.get('chi_b', 1)
        self._tau_b = self.expe.get('tau_b', 100)
        self._kappa_b = self.expe.get('kappa_b', 0.9)
        self._update_gstep_theta()
        self._update_gstep_phi()
        self._update_gstep_y()


    def _random_ss_init(self):
        ''' Sufficient Statistics Initialization '''
        K = self._len['K']
        N = self._len['N']
        nnz = self._len['nnz']
        nnzsum = self._len['nnzsum']
        dims = self._len['dims']

        self.N_theta_left = (dims[:, None] * np.random.dirichlet([0.5]*K, N))
        self.N_theta_right = (dims[:, None] * np.random.dirichlet([0.5]*K, N))

        self.N_phi = np.random.dirichlet([0.5]*K**2).reshape(K,K) *nnz

        #self.N_Y = np.random.poisson(0.1, (K,K)) * N
        self.N_Y = np.random.dirichlet([0.1]*K**2).reshape(K,K) * nnzsum

        if self._is_symmetric:
            self.N_theta_left = self.N_theta_right
            self.N_phi = np.triu(self.N_phi) + np.triu(self.N_phi, 1).T
            self.N_Y = np.triu(self.N_Y) + np.triu(self.N_Y, 1).T


    def _update_gstep_theta(self):
        ''' Gradient converge for kappa _in (0.5,1] '''
        chi = self._chi_a
        tau = self._tau_a
        kappa = self._kappa_a

        self.gstep_theta = chi / ((tau + self._timestep_a)**kappa)

    def _update_gstep_phi(self):
        chi = self._chi_b
        tau = self._tau_b
        kappa = self._kappa_b

        self.gstep_phi =  chi / ((tau + self._timestep_b)**kappa)

    def _update_gstep_y(self):
        chi = self._chi_b
        tau = self._tau_b
        kappa = self._kappa_b

        self.gstep_y =  chi / ((tau + self._timestep_c)**kappa)


    def _reduce_latent(self):
        theta = self.N_theta_right + self.N_theta_left + np.tile(self.hyper_theta, (self.N_theta_left.shape[0],1))
        self._theta = (theta.T / theta.sum(axis=1)).T

        k = self.N_Y + self.hyper_phi[0]
        p = (self.N_phi + self.hyper_phi[1] + 1)**-1
        self._phi = lambda x:sp.stats.nbinom.pmf(x, k, 1-p)
        #mean = k*p / (1-p)
        #var = k*p / (1-p)**2

        return self._theta, self._phi

    def _reduce_one(self, i, j, xij, update_kernel=True):

        if self._is_symmetric:
            self.pik = self.pjk = self.N_theta_left[i] + self.hyper_theta
            self.pjk = self.pik
        else:
            self.pik = self.N_theta_left[i] + self.hyper_theta
            self.pjk = self.N_theta_right[j] + self.hyper_theta

        if update_kernel:
            k = self.N_Y + self.hyper_phi[0]
            p = (self.N_phi + self.hyper_phi[1] + 1)**-1
            # @debug: Some invalie values here sometime !!
            self._kernel = lambda x:sp.stats.nbinom.pmf(x, k, 1-p)
            self._lut_nbinom = [sp.stats.nbinom.pmf(x, k, 1-p) for x in range(42)]
            #kernel = sp.stats.nbinom.pmf(xij, k, 1-p)

        if len(self._lut_nbinom) > xij:
            # Wins some times...
            kernel = self._lut_nbinom[xij]
        else:
            kernel = self._kernel(xij)

        # debug: Underflow
        kernel[kernel<=1e-300] = 1e-300

        outer_kk = np.log(np.outer(self.pik, self.pjk)) + np.log(kernel)

        return lognormalize(outer_kk.ravel())

    def _optimize_hyper(self):
        n, t = self.hyper_phi
        K = self.N_theta_left.shape[1]

        phi_mean = self.get_mean_phi() / (1+self._is_symmetric) * self._len['nnz_ones']
        # debug: Underflow
        #phi_mean[phi_mean<=1e-300] = 1e-300
        phi_sum = phi_mean.sum()
        phi_mean[phi_mean<=1e-300] = 1e-300
        log_phi_sum = np.log(phi_mean).sum()

        K_len = (K*(K-1)/2 +K) if self._is_symmetric else K**2

        t = n * K_len / phi_sum

        # http://bariskurt.com/calculating-the-inverse-of-digamma-function/
        #s = log_phi_sum / K_len + np.log(t)
        #x0 = lambda x: -1/(x + sp.special.digamma(1)) if x<-2.22 else np.exp(x)+0.5

        s = -log_phi_sum / K_len - np.log(K_len/phi_sum)
        x0 = lambda x: (3 - x + np.sqrt((x-3)**2 + 24*x))/(12*x)
        dig = lambda x: np.log(x) - sp.special.digamma(x) -s
        n = float(sp.optimize.minimize(dig, x0(s)).x)

        self.hyper_phi = np.array([n,t])

    def get_mean_phi(self):
        k = self.N_Y + self.hyper_phi[0]
        p = (self.N_phi + self.hyper_phi[1] + 1)**-1
        return  k*p / (1-p)

    def get_nb_ss(self):
        k = self.N_Y + self.hyper_phi[0]
        p = (self.N_phi + self.hyper_phi[1] + 1)**-1
        mean = k*p / (1-p)
        var = k*p / (1-p)**2
        return mean, var

    def likelihood(self, theta=None, phi=None):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi

        #lut_nbinom = [phi(x) for x in range(42)]
        kern = self.get_mean_phi()

        likelihood = []

        for i,j, xij in self.data_test:

            # 1ts interpretation
            #if len(lut_nbinom) > xij:
            #    # Wins some times...
            #    kernel = lut_nbinom[xij]
            #else:
            #    kernel = phi(xij)
            #l = theta[i].dot(kernel).dot(theta[j])

            # 2nd interpretation
            qij = theta[i].dot(kern).dot(theta[j])
            l = sp.stats.poisson.pmf(xij, qij)

            likelihood.append(l or ma.masked)



        likelihood = ma.array(likelihood)

        return likelihood

    def compute_entropy(self, theta=None, phi=None):
        if theta is None:
            theta, phi = self._reduce_latent()
        return self.compute_entropy_t(theta, phi)

    def compute_entropy_t(self, theta=None, phi=None):
        if theta is None:
            theta, phi = self._reduce_latent()

        pij = self.likelihood(theta, phi)

        # Log-likelihood
        ll = np.log(pij).sum()

        # Entropy
        entropy_t = ll
        #self._entropy_t = - ll / self._len['nnz_t']

        # Perplexity is 2**H(X).

        return entropy_t

    def compute_perplexity(self, theta=None, phi=None):
        if theta is None:
            theta, phi = self._reduce_latent()

        entropy = self.compute_entropy(theta,phi)
        nnz = self.data_test.shape[0]
        return 2 ** (-entropy / nnz)


    def compute_elbo(self, theta=None, phi=None):
        # how to compute elbo for all possible links weights, mean?
        return None

    def compute_roc(self, theta=None, phi=None, treshold='auto'):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        if theta is None:
            theta, phi = self._reduce_latent()

        weights = self.data_test[:,2]
        phi_mean = self.get_mean_phi()

        treshold = 'mean_model'
        if treshold == 'mean_data':
            mean = weights[weights>0].mean()
            std = weights[weights>0].std()
        if treshold == 'mean_model':
            mean = phi_mean.mean()
            std = phi_mean.std()

        #trsh = int(mean_w)
        trsh = int(mean+std)

        y_true = []
        probas = []
        for i,j, xij in self.data_test:
            qij = theta[i].dot(phi_mean).dot(theta[j])
            #qij = theta[i].dot(phi_mean).dot(theta[j])
            #print(qij, xij)
            pij =  1 - sp.stats.poisson.cdf(trsh, qij)
            y_true.append(xij>0)
            probas.append(pij)


        fpr, tpr, thresholds = roc_curve(y_true, probas)

        roc = auc(fpr, tpr)
        return roc

    def compute_wsim(self, theta=None, phi=None):
        if theta is None:
            theta, phi = self._reduce_latent()

        # NB mean/var
        mean, var = self.get_nb_ss()
        phi = mean

        weights = self.data_test[:,2]
        nnz = self.data_test.shape[0]
        sim = []
        for i,j,_w in self.data_test:
            w = np.random.poisson(theta[i].dot(phi).dot(theta[j].T))
            sim.append(w)

        # l1 norm
        mean_dist = np.abs(np.array(w) - weights.T).sum() / nnz
        return mean_dist


    def compute_measures(self, begin_it=0):
        ''' Compute measure as model attributes.
            begin_it: is the time of the begining of the iteration.
        '''

        if self.expe.get('deactivate_measures'):
            return


        params = self._reduce_latent()
        self.time_it = (time() - begin_it)

        for meas in self.expe._csv_typo.split():

            if meas.lstrip('_') == 'entropy':
                old_entropy = self._entropy
                _meas = self.compute_entropy(*params)
                self.entropy_diff = _meas - old_entropy
            elif hasattr(self, 'compute_'+meas.lstrip('_')):
                _meas = getattr(self, 'compute_'+meas.lstrip('_'))(*params)
            else:
                # Assume already computed
                getattr(self, meas) # raise exception if not here.
                continue

            setattr(self, meas, _meas)

        return



    def update_hyper(self, hyper):
        pass

    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        #self.update_hyper(hyperparams)
        #alpha, gmma, delta = self.get_hyper()

        # predictive
        try: theta, phi = self.get_params()
        except: return self.generate(N, K, hyperparams, 'generative', symmetric)
        K = theta.shape[1]

        raise NotImplementedError

        pij = self.likelihood(theta, phi)
        pij = np.clip(pij, 0, 1)
        Y = sp.stats.bernoulli.rvs(pij)

        return Y

    def fit(self, frontend):
        ''' chunk is the number of row to threat in a minibach '''

        self._init(frontend)
        mnb_total = frontend.num_mnb()

        # Init sampling variables
        observed_pt = 0
        mnb_num = 0
        vertex = None

        self._entropy = self.compute_entropy()
        print( '__init__ Entropy: %f' % self._entropy)
        for _it, obj in enumerate(frontend):

            source, target, weight = obj
            if type(source) is str:
                #print(_it, source, target, weight)
                _set_pos = source
                _vertex = target['vertex']
                _direction = target['direction']
                _scaler = weight
                _qij_samples = []
                _node_idxs = []
                _weights = []
                new_mnb = True

                update_kernel = True
            else:
                i = source
                j = target
                weights.append(weight)
                if direction == 0:
                    node_idxs.append(j)
                else:
                    node_idxs.append(i)

                # Maximization
                qij_samples.append( self._reduce_one(i,j, weight, update_kernel).reshape(self._len['K'], self._len['K']) )

                # Update local gradient / Expectation
                # will be longer but faster ???
                # try updating only  J => ima sure it will increase perf (j here and i in the global set (ore reversly selon la direction)
                #
                #self.N_theta_left[i] = (1 - self.gstep_theta)*self.N_theta_left[i] + (self.gstep_theta * scaler * qij_samples[-1].sum(0))
                #self.N_theta_right[j] = (1 - self.gstep_theta)*self.N_theta_right[j] + (self.gstep_theta * scaler * qij_samples[-1].sum(1))
                #self._timestep_a += 1
                #self._update_gstep_theta()
                update_kernel = False

                observed_pt += 1

            if vertex is None or not qij_samples:
                # Enter here only once !%!
                set_pos = _set_pos
                vertex = _vertex
                direction = _direction
                scaler = _scaler
                qij_samples = _qij_samples
                node_idxs = _node_idxs
                weights = _weights
                new_mnb = False
                continue

            if new_mnb:

                qijs = np.asarray(qij_samples)
                # Update global gradient / Expectation
                # Normalize or not ?
                #norm=1
                norm = qijs.shape[0]
                qijs_sum = qijs.sum(0)

                if direction == 0:
                    self.N_theta_left[i] = (1-self.gstep_theta)*self.N_theta_left[i] + self.gstep_theta*scaler*qijs_sum.sum(0) /norm
                    self.N_theta_right[node_idxs] = (1-self.gstep_theta)*self.N_theta_right[node_idxs] + self.gstep_theta*scaler*qijs.sum(2) /norm
                else:
                    self.N_theta_left[node_idxs] = (1-self.gstep_theta)*self.N_theta_left[node_idxs] + self.gstep_theta*scaler*qijs.sum(1) /norm
                    self.N_theta_right[j] = (1-self.gstep_theta)*self.N_theta_right[j] + self.gstep_theta*scaler*qijs_sum.sum(1) /norm

                self._timestep_a += scaler
                self._update_gstep_theta()

                #scaler = self._len['nnz']
                self.N_phi = (1 - self.gstep_phi)*self.N_phi + self.gstep_phi * scaler * qijs_sum /norm
                if set_pos != '0':
                    self.N_Y = (1 - self.gstep_y)*self.N_Y + self.gstep_y * scaler * np.sum([weights[n]*qijs[n] for n in range(len(weights)) ],0) /norm
                    self._timestep_c += scaler
                    self._update_gstep_y()

                    if self._hyper_phi == 'auto':
                        self._optimize_hyper()

                self._timestep_b += scaler
                self._update_gstep_phi()


                # Allocate current state variable
                set_pos = _set_pos
                vertex = _vertex
                direction = _direction
                scaler = _scaler
                qij_samples = _qij_samples
                node_idxs = _node_idxs
                weights = _weights
                new_mnb = False

                mnb_num += 1

                if mnb_num % 50 == 0:
                    prop_edge = observed_pt / self._len['nnz']
                    self._observed_pt = observed_pt
                    self.compute_measures()

                    print('.', end='')
                    self.log.info('it %d | prop edge: %.2f | mnb %d/%d, %s, Entropy: %f,  diff: %f' % (_it, prop_edge,
                                                                                                       mnb_num, mnb_total,
                                                                                                       '/'.join((self.expe.model, self.expe.corpus)),
                                                                                                       self._entropy, self.entropy_diff))

                    if self.expe.get('_write'):
                        self.write_current_state(self)
                        if mnb_num % 2000 == 0:
                            self.save(silent=True)
                            sys.stdout.flush()

